#line 2 "lexer.cc"

#line 4 "lexer.cc"

#define  YY_INT_ALIGNED short int

/* A lexical scanner generated by flex */

/* %not-for-header */
/* %if-c-only */
/* %if-not-reentrant */

/* %endif */
/* %endif */
/* %ok-for-header */

#define FLEX_SCANNER
#define YY_FLEX_MAJOR_VERSION 2
#define YY_FLEX_MINOR_VERSION 6
#define YY_FLEX_SUBMINOR_VERSION 4
#if YY_FLEX_SUBMINOR_VERSION > 0
#define FLEX_BETA
#endif

/* %if-c++-only */
/* %endif */

/* %if-c-only */

/* %endif */

/* %if-c-only */

/* %endif */

/* First, we deal with  platform-specific or compiler-specific issues. */

/* begin standard C headers. */
/* %if-c-only */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
/* %endif */

/* %if-tables-serialization */
/* %endif */
/* end standard C headers. */

/* %if-c-or-c++ */
/* flex integer type definitions */

#ifndef FLEXINT_H
#define FLEXINT_H

/* C99 systems have <inttypes.h>. Non-C99 systems may or may not. */

#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L

/* C99 says to define __STDC_LIMIT_MACROS before including stdint.h,
 * if you want the limit (max/min) macros for int types. 
 */
#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS 1
#endif

#include <inttypes.h>
typedef int8_t flex_int8_t;
typedef uint8_t flex_uint8_t;
typedef int16_t flex_int16_t;
typedef uint16_t flex_uint16_t;
typedef int32_t flex_int32_t;
typedef uint32_t flex_uint32_t;
#else
typedef signed char flex_int8_t;
typedef short int flex_int16_t;
typedef int flex_int32_t;
typedef unsigned char flex_uint8_t; 
typedef unsigned short int flex_uint16_t;
typedef unsigned int flex_uint32_t;

/* Limits of integral types. */
#ifndef INT8_MIN
#define INT8_MIN               (-128)
#endif
#ifndef INT16_MIN
#define INT16_MIN              (-32767-1)
#endif
#ifndef INT32_MIN
#define INT32_MIN              (-2147483647-1)
#endif
#ifndef INT8_MAX
#define INT8_MAX               (127)
#endif
#ifndef INT16_MAX
#define INT16_MAX              (32767)
#endif
#ifndef INT32_MAX
#define INT32_MAX              (2147483647)
#endif
#ifndef UINT8_MAX
#define UINT8_MAX              (255U)
#endif
#ifndef UINT16_MAX
#define UINT16_MAX             (65535U)
#endif
#ifndef UINT32_MAX
#define UINT32_MAX             (4294967295U)
#endif

#ifndef SIZE_MAX
#define SIZE_MAX               (~(size_t)0)
#endif

#endif /* ! C99 */

#endif /* ! FLEXINT_H */

/* %endif */

/* begin standard C++ headers. */
/* %if-c++-only */
/* %endif */

/* TODO: this is always defined, so inline it */
#define yyconst const

#if defined(__GNUC__) && __GNUC__ >= 3
#define yynoreturn __attribute__((__noreturn__))
#else
#define yynoreturn
#endif

/* %not-for-header */
/* Returned upon end-of-file. */
#define YY_NULL 0
/* %ok-for-header */

/* %not-for-header */
/* Promotes a possibly negative, possibly signed char to an
 *   integer in range [0..255] for use as an array index.
 */
#define YY_SC_TO_UI(c) ((YY_CHAR) (c))
/* %ok-for-header */

/* %if-reentrant */
/* %endif */

/* %if-not-reentrant */

/* %endif */

/* Enter a start condition.  This macro really ought to take a parameter,
 * but we do it the disgusting crufty way forced on us by the ()-less
 * definition of BEGIN.
 */
#define BEGIN (yy_start) = 1 + 2 *
/* Translate the current start state into a value that can be later handed
 * to BEGIN to return to the state.  The YYSTATE alias is for lex
 * compatibility.
 */
#define YY_START (((yy_start) - 1) / 2)
#define YYSTATE YY_START
/* Action number for EOF rule of a given start state. */
#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)
/* Special action meaning "start processing a new file". */
#define YY_NEW_FILE yyrestart( yyin  )
#define YY_END_OF_BUFFER_CHAR 0

/* Size of default input buffer. */
#ifndef YY_BUF_SIZE
#ifdef __ia64__
/* On IA-64, the buffer size is 16k, not 8k.
 * Moreover, YY_BUF_SIZE is 2*YY_READ_BUF_SIZE in the general case.
 * Ditto for the __ia64__ case accordingly.
 */
#define YY_BUF_SIZE 32768
#else
#define YY_BUF_SIZE 16384
#endif /* __ia64__ */
#endif

/* The state buf must be large enough to hold one state per character in the main buffer.
 */
#define YY_STATE_BUF_SIZE   ((YY_BUF_SIZE + 2) * sizeof(yy_state_type))

#ifndef YY_TYPEDEF_YY_BUFFER_STATE
#define YY_TYPEDEF_YY_BUFFER_STATE
typedef struct yy_buffer_state *YY_BUFFER_STATE;
#endif

#ifndef YY_TYPEDEF_YY_SIZE_T
#define YY_TYPEDEF_YY_SIZE_T
typedef size_t yy_size_t;
#endif

/* %if-not-reentrant */
extern int yyleng;
/* %endif */

/* %if-c-only */
/* %if-not-reentrant */
extern FILE *yyin, *yyout;
/* %endif */
/* %endif */

#define EOB_ACT_CONTINUE_SCAN 0
#define EOB_ACT_END_OF_FILE 1
#define EOB_ACT_LAST_MATCH 2
    
    #define YY_LESS_LINENO(n)
    #define YY_LINENO_REWIND_TO(ptr)
    
/* Return all but the first "n" matched characters back to the input stream. */
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		*yy_cp = (yy_hold_char); \
		YY_RESTORE_YY_MORE_OFFSET \
		(yy_c_buf_p) = yy_cp = yy_bp + yyless_macro_arg - YY_MORE_ADJ; \
		YY_DO_BEFORE_ACTION; /* set up yytext again */ \
		} \
	while ( 0 )
#define unput(c) yyunput( c, (yytext_ptr)  )

#ifndef YY_STRUCT_YY_BUFFER_STATE
#define YY_STRUCT_YY_BUFFER_STATE
struct yy_buffer_state
	{
/* %if-c-only */
	FILE *yy_input_file;
/* %endif */

/* %if-c++-only */
/* %endif */

	char *yy_ch_buf;		/* input buffer */
	char *yy_buf_pos;		/* current position in input buffer */

	/* Size of input buffer in bytes, not including room for EOB
	 * characters.
	 */
	int yy_buf_size;

	/* Number of characters read into yy_ch_buf, not including EOB
	 * characters.
	 */
	int yy_n_chars;

	/* Whether we "own" the buffer - i.e., we know we created it,
	 * and can realloc() it to grow it, and should free() it to
	 * delete it.
	 */
	int yy_is_our_buffer;

	/* Whether this is an "interactive" input source; if so, and
	 * if we're using stdio for input, then we want to use getc()
	 * instead of fread(), to make sure we stop fetching input after
	 * each newline.
	 */
	int yy_is_interactive;

	/* Whether we're considered to be at the beginning of a line.
	 * If so, '^' rules will be active on the next match, otherwise
	 * not.
	 */
	int yy_at_bol;

    int yy_bs_lineno; /**< The line count. */
    int yy_bs_column; /**< The column count. */

	/* Whether to try to fill the input buffer when we reach the
	 * end of it.
	 */
	int yy_fill_buffer;

	int yy_buffer_status;

#define YY_BUFFER_NEW 0
#define YY_BUFFER_NORMAL 1
	/* When an EOF's been seen but there's still some text to process
	 * then we mark the buffer as YY_EOF_PENDING, to indicate that we
	 * shouldn't try reading from the input source any more.  We might
	 * still have a bunch of tokens to match, though, because of
	 * possible backing-up.
	 *
	 * When we actually see the EOF, we change the status to "new"
	 * (via yyrestart()), so that the user can continue scanning by
	 * just pointing yyin at a new input file.
	 */
#define YY_BUFFER_EOF_PENDING 2

	};
#endif /* !YY_STRUCT_YY_BUFFER_STATE */

/* %if-c-only Standard (non-C++) definition */
/* %not-for-header */
/* %if-not-reentrant */

/* Stack of input buffers. */
static size_t yy_buffer_stack_top = 0; /**< index of top of stack. */
static size_t yy_buffer_stack_max = 0; /**< capacity of stack. */
static YY_BUFFER_STATE * yy_buffer_stack = NULL; /**< Stack as an array. */
/* %endif */
/* %ok-for-header */

/* %endif */

/* We provide macros for accessing buffer states in case in the
 * future we want to put the buffer states in a more general
 * "scanner state".
 *
 * Returns the top of the stack, or NULL.
 */
#define YY_CURRENT_BUFFER ( (yy_buffer_stack) \
                          ? (yy_buffer_stack)[(yy_buffer_stack_top)] \
                          : NULL)
/* Same as previous macro, but useful when we know that the buffer stack is not
 * NULL or when we need an lvalue. For internal use only.
 */
#define YY_CURRENT_BUFFER_LVALUE (yy_buffer_stack)[(yy_buffer_stack_top)]

/* %if-c-only Standard (non-C++) definition */

/* %if-not-reentrant */
/* %not-for-header */
/* yy_hold_char holds the character lost when yytext is formed. */
static char yy_hold_char;
static int yy_n_chars;		/* number of characters read into yy_ch_buf */
int yyleng;

/* Points to current character in buffer. */
static char *yy_c_buf_p = NULL;
static int yy_init = 0;		/* whether we need to initialize */
static int yy_start = 0;	/* start state number */

/* Flag which is used to allow yywrap()'s to do buffer switches
 * instead of setting up a fresh yyin.  A bit of a hack ...
 */
static int yy_did_buffer_switch_on_eof;
/* %ok-for-header */

/* %endif */

void yyrestart ( FILE *input_file  );
void yy_switch_to_buffer ( YY_BUFFER_STATE new_buffer  );
YY_BUFFER_STATE yy_create_buffer ( FILE *file, int size  );
void yy_delete_buffer ( YY_BUFFER_STATE b  );
void yy_flush_buffer ( YY_BUFFER_STATE b  );
void yypush_buffer_state ( YY_BUFFER_STATE new_buffer  );
void yypop_buffer_state ( void );

static void yyensure_buffer_stack ( void );
static void yy_load_buffer_state ( void );
static void yy_init_buffer ( YY_BUFFER_STATE b, FILE *file  );
#define YY_FLUSH_BUFFER yy_flush_buffer( YY_CURRENT_BUFFER )

YY_BUFFER_STATE yy_scan_buffer ( char *base, yy_size_t size  );
YY_BUFFER_STATE yy_scan_string ( const char *yy_str  );
YY_BUFFER_STATE yy_scan_bytes ( const char *bytes, int len  );

/* %endif */

void *yyalloc ( yy_size_t  );
void *yyrealloc ( void *, yy_size_t  );
void yyfree ( void *  );

#define yy_new_buffer yy_create_buffer
#define yy_set_interactive(is_interactive) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){ \
        yyensure_buffer_stack (); \
		YY_CURRENT_BUFFER_LVALUE =    \
            yy_create_buffer( yyin, YY_BUF_SIZE ); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_is_interactive = is_interactive; \
	}
#define yy_set_bol(at_bol) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){\
        yyensure_buffer_stack (); \
		YY_CURRENT_BUFFER_LVALUE =    \
            yy_create_buffer( yyin, YY_BUF_SIZE ); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_at_bol = at_bol; \
	}
#define YY_AT_BOL() (YY_CURRENT_BUFFER_LVALUE->yy_at_bol)

/* %% [1.0] yytext/yyin/yyout/yy_state_type/yylineno etc. def's & init go here */
/* Begin user sect3 */

#define yywrap() (/*CONSTCOND*/1)
#define YY_SKIP_YYWRAP

#define FLEX_DEBUG
typedef flex_uint8_t YY_CHAR;

FILE *yyin = NULL, *yyout = NULL;

typedef int yy_state_type;

extern int yylineno;
int yylineno = 1;

extern char *yytext;
#ifdef yytext_ptr
#undef yytext_ptr
#endif
#define yytext_ptr yytext

/* %% [1.5] DFA */

/* %if-c-only Standard (non-C++) definition */

static yy_state_type yy_get_previous_state ( void );
static yy_state_type yy_try_NUL_trans ( yy_state_type current_state  );
static int yy_get_next_buffer ( void );
static void yynoreturn yy_fatal_error ( const char* msg  );

/* %endif */

/* Done after the current pattern has been matched and before the
 * corresponding action - sets up yytext.
 */
#define YY_DO_BEFORE_ACTION \
	(yytext_ptr) = yy_bp; \
/* %% [2.0] code to fiddle yytext and yyleng for yymore() goes here \ */\
	yyleng = (int) (yy_cp - yy_bp); \
	(yy_hold_char) = *yy_cp; \
	*yy_cp = '\0'; \
/* %% [3.0] code to copy yytext_ptr to yytext[] goes here, if %array \ */\
	(yy_c_buf_p) = yy_cp;
/* %% [4.0] data tables for the DFA and the user's section 1 definitions go here */
#define YY_NUM_RULES 55
#define YY_END_OF_BUFFER 56
/* This struct is not used in this scanner,
   but its presence is necessary. */
struct yy_trans_info
	{
	flex_int32_t yy_verify;
	flex_int32_t yy_nxt;
	};
static const flex_int16_t yy_accept[236] =
    {   0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    7,    7,    7,    7,    0,    0,    0,    0,
        0,    0,    0,    0,   56,   39,   38,   54,    1,    2,
       23,   42,   54,   42,   54,   27,   28,   31,   42,   48,
       48,   45,   32,   29,   30,   38,   33,   34,   35,   20,
        1,    2,   17,   20,   20,   20,   21,    1,    2,   21,
       21,   21,   19,    1,    2,   19,   19,   19,   19,   22,
        1,    2,   55,   22,   22,   22,   15,    1,    2,   14,
       15,   15,   15,    7,    1,    2,    9,    8,    7,    7,
        8,   39,   54,   54,   34,   39,   39,   39,    0,   38,

       38,   38,    1,   42,   41,   41,    0,    0,    0,    0,
       53,    0,   42,    0,    0,    0,    0,    0,   42,   41,
        5,   47,    0,   48,    0,    0,    0,    0,   45,   38,
       38,    0,    0,   16,    0,   18,    7,    1,    4,    0,
        3,   39,   39,   39,    0,    0,   37,   36,   43,   44,
       40,    0,    0,    0,    0,    0,   53,    0,    0,   26,
       52,    0,    0,    0,    0,    0,    0,   52,    0,    0,
       11,   12,   11,   11,   42,   11,   46,    0,    0,   51,
        0,   51,   49,    0,   50,    0,   38,   38,   38,   38,
       10,    0,    0,    6,   37,   37,   37,   37,   36,   36,

       36,   43,   44,   52,   11,   13,   11,   51,    0,    0,
       51,    0,   49,    0,   50,    0,   38,   25,   25,   25,
        0,   37,   37,   37,    0,   51,   38,   38,    0,   51,
        0,   24,   24,   24,    0
    } ;

static const YY_CHAR yy_ec[256] =
    {   0,
        1,    2,    3,    4,    5,    6,    7,    4,    8,    9,
       10,   11,   11,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,   12,   13,   14,   15,   13,   13,   13,   16,   17,
       18,   13,   19,   20,   21,   22,   13,   23,   24,   24,
       24,   24,   24,   24,   24,   25,   25,   26,   27,   13,
       13,   13,   13,   13,   28,   28,   28,   28,   29,   28,
       30,   30,   30,   30,   30,   30,   30,   30,   31,   30,
       30,   30,   30,   30,   30,   30,   30,   32,   30,   30,
       33,   34,   35,   13,   36,   37,   38,   38,   39,   39,

       40,   38,   41,   41,   41,   41,   41,   41,   41,   42,
       43,   41,   41,   42,   41,   42,   41,   42,   41,   44,
       41,   41,   45,   46,   47,   13,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,

        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4,    4,    4,    4,    4,    4,
        4,    4,    4,    4,    4
    } ;

static const YY_CHAR yy_meta[48] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    2,    1,
        1,    1,    3,    4,    5,    4,    4,    1,    3,    1,
        6,    3,    7,    7,    7,    3,    1,    7,    7,    4,
        4,    4,    4,    3,    1,    7,    1,    7,    7,    7,
        4,    4,    4,    4,    4,    3,    1
    } ;

static const flex_int16_t yy_base[256] =
    {   0,
        0,    0,   43,   86,  129,  172,  215,  258,  301,  344,
      387,  430,  473,  516,  559,  602,  649,    0,  691,  734,
      777,  820,  863,  906,  283,  953,  997, 1844,   44, 1844,
     1844, 1029, 1075, 1110, 1156, 1844, 1844, 1844,   36, 1189,
       31,    0, 1844, 1844, 1844,   42, 1844,  254, 1844, 1844,
       67, 1844, 1844,   29,  253, 1844, 1844,   75, 1844,   37,
      244, 1844, 1844,   87, 1844,   61,  243,    0, 1844, 1844,
       99, 1844, 1844,   65,  238, 1844, 1844,  105, 1844, 1844,
       72,  233, 1844,    0,  108,    0, 1844,   79,  227,    0,
       80, 1233,  225,   81,  219,    0,    0,    0, 1277,    0,

        0,    0,  111,  218, 1844,  214,    0,    0,    0,    0,
     1844, 1323,   82,  217,  206,  200,  196, 1370, 1417,  120,
     1844, 1844,   45,    0,  128,  131,  175,  135,    0,  163,
      166,  189,  157, 1844,  186, 1844,    0,  120, 1844,    0,
     1844,    0,    0,    0, 1464,  153, 1508, 1552,    0,    0,
     1844,    0,    0,    0,    0,    0,  182,    0,    0, 1844,
     1844,  167,  163,  157,  156,  154,  152,  149,  141,  132,
        0, 1844,  137,    0,  130,  119, 1844,  182,  185,  205,
      138,  190,  208,  214,   86,   78,    0,  228,   84,    0,
     1844, 1596,  115, 1844, 1640,    0,    0,    0,    0,    0,

        0,    0,    0,   89,    0, 1844,    0,  248,  257,  262,
      233,  271,  219,  276,   53,   45,  291,  314,    0,    0,
        0,    0,    0,    0,  171,  305,  308,  311, 1684,  314,
      334,  357,    0,    0, 1844, 1728, 1735, 1742, 1749, 1756,
     1763, 1770, 1775, 1780, 1786, 1793, 1800,   66, 1807, 1814,
     1821, 1826, 1830, 1836,   64
    } ;

static const flex_int16_t yy_def[256] =
    {   0,
      235,    1,  236,  236,  237,  237,  238,  238,  239,  239,
      240,  240,  241,  241,  241,  241,    1,   17,  242,  242,
      242,  242,  242,  242,  235,  235,  235,  235,  235,  235,
      235,  243,  235,  243,  235,  235,  235,  235,   34,  235,
       40,  244,  235,  235,  235,   27,  235,  235,  235,  235,
      235,  235,  235,  235,  245,  235,  235,  235,  235,  235,
      245,  235,  235,  235,  235,  235,  245,  246,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  247,  247,  247,  235,  235,  247,  247,
      235,  235,  235,  235,  235,   26,   26,   26,  235,   27,

       27,   27,  235,   34,  235,  235,   33,   33,   33,   33,
      235,  235,   34,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,   41,  235,  235,  248,  235,  244,   27,
       27,  249,  250,  235,  249,  235,  247,  247,  235,  251,
      235,   92,   92,   92,  235,  235,  235,  235,  252,  253,
      235,   33,   33,   33,   33,   33,   33,  112,   33,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      254,  235,  254,  254,  119,  254,  235,  235,  235,  235,
      235,  235,  235,  235,  255,  248,  130,   27,  130,  188,
      235,  235,  250,  235,  235,  147,  147,  147,  148,  148,

      148,  252,  253,  235,  254,  235,  254,  235,  235,  235,
      235,  235,  235,  235,  255,  255,   27,  192,  218,  218,
      192,  195,  195,  195,  235,  235,   27,   27,  235,  235,
      235,  229,  232,  232,    0,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235
    } ;

static const flex_int16_t yy_nxt[1892] =
    {   0,
       26,   27,   28,   28,   29,   28,   28,   30,   31,   29,
       28,   29,   32,   33,   34,   35,   36,   37,   32,   38,
       39,   32,   40,   41,   41,   42,   43,   26,   26,   26,
       26,   26,   44,   32,   45,   46,   47,   27,   27,   27,
       27,   27,   27,   27,   48,   32,   49,   51,  103,  133,
       52,   53,   51,  103,   51,  103,  119,  133,  120,  120,
      120,  235,  235,   54,  130,  130,  130,  180,  180,  180,
      215,  103,  185,  235,  235,  121,  103,  131,  103,  103,
      216,  133,  121,  121,  103,  133,  103,   55,  216,   56,
       51,  103,  133,   52,   53,   51,  103,   51,  103,  133,

      133,  133,  104,  103,  161,  235,   54,  121,  103,  103,
      103,  121,  138,  186,  103,  103,  103,  138,  121,  138,
      103,  216,  103,  172,  138,  139,  141,  121,  160,  138,
       55,  138,   56,   58,  177,  193,   59,   53,   58,  207,
       58,  171,  178,  178,  178,  206,  181,  161,  181,   60,
      182,  182,  182,  183,  183,  179,  161,  124,  124,  124,
      182,  182,  182,  125,  161,  172,  184,  204,  173,  161,
      128,  161,  161,   61,  125,   62,   58,  193,  161,   59,
       53,   58,  161,   58,  123,  187,  187,  187,  130,  130,
      130,  188,   60,  226,  226,  226,  177,  153,  189,  160,

      235,  131,  190,  192,  178,  178,  178,  178,  178,  178,
      186,  161,  211,  211,  211,  161,   61,  179,   62,   64,
      179,  161,   65,   53,   64,  212,   64,  208,  208,  208,
      213,  213,  161,  209,  151,   66,  183,  183,  104,  135,
      210,  213,  213,  214,  209,  146,  181,  140,  181,  184,
      217,  217,  217,  135,  214,  211,  211,  211,  135,   67,
       68,   69,   64,  135,  135,   65,   53,   64,  212,   64,
      208,  208,  208,  135,  132,  225,  209,  225,   66,  226,
      226,  226,  235,  210,  208,  208,  208,  209,  235,  235,
      209,  235,  235,  211,  211,  211,  235,  210,  213,  213,

      235,  209,   67,   68,   69,   71,  212,  235,   72,   73,
       71,  214,   71,  227,  227,  227,  235,  235,  229,  235,
      235,   74,  229,  229,  229,  229,  228,  230,  230,  230,
      227,  227,  227,  227,  227,  227,  230,  230,  230,  235,
      231,  235,  235,  228,  235,   75,  228,   76,   71,  231,
      235,   72,   73,   71,  235,   71,  230,  230,  230,  235,
      235,  235,  235,  235,   74,  235,  235,  235,  235,  231,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,   75,  235,
       76,   78,  235,  235,   79,   80,   78,  235,   78,  235,

      235,  235,  235,  235,  235,  235,  235,   81,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,   82,  235,   83,   78,  235,  235,   79,   80,   78,
      235,   78,  235,  235,  235,  235,  235,  235,  235,  235,
       81,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,   82,  235,   83,   85,  235,  235,
       86,   87,   85,  235,   85,  235,  235,  235,  235,  235,
      235,  235,  235,   88,  235,  235,  235,  235,  235,  235,

      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,   89,  235,   90,
       85,  235,  235,   86,   87,   85,  235,   85,  235,  235,
      235,  235,  235,  235,  235,  235,   88,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
       89,  235,   90,   85,  235,  235,   86,   87,   85,  235,
       85,  235,  235,  235,  235,  235,  235,  235,  235,   91,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,

      235,  235,  235,   89,  235,   90,   85,  235,  235,   86,
       87,   85,  235,   85,  235,  235,  235,  235,  235,  235,
      235,  235,   91,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,   89,  235,   90,   92,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,   28,   28,   93,   28,  235,  235,   28,  235,   94,
       28,   28,   28,   28,   28,  235,   92,   92,   92,   92,
       92,  235,   28,  235,   27,  235,  235,  235,  235,  235,
      235,  235,  235,  235,   28,   29,  235,  235,   30,   73,

       29,  235,   29,  235,  235,  235,  235,  235,  235,  235,
      235,   94,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,   95,  235,   49,   29,  235,
      235,   30,   73,   29,  235,   29,  235,  235,  235,  235,
      235,  235,  235,  235,   94,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,   95,  235,
       49,   29,  235,  235,   30,   73,   29,  235,   29,  235,
      235,  235,  235,  235,  235,  235,  235,   94,  235,  235,

      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,   95,  235,   49,   29,  235,  235,   30,   73,   29,
      235,   29,  235,  235,  235,  235,  235,  235,  235,  235,
       94,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,   95,  235,   49,   29,  235,  235,
       30,   73,   29,  235,   29,  235,  235,  235,  235,  235,
      235,  235,  235,   94,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,

      235,  235,  235,  235,  235,  235,  235,   95,  235,   49,
       29,  235,  235,   30,   73,   29,  235,   29,  235,  235,
      235,  235,  235,  235,  235,  235,   94,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
       95,  235,   49,   96,   97,   98,  235,  235,  235,   97,
      235,  235,  235,  235,  235,  235,  235,  235,   97,  235,
      235,  235,  235,  235,   99,   98,   98,   98,  235,  235,
       96,   96,   96,   96,   96,  235,  235,  235,   97,  235,
       97,   97,   97,   97,   97,   97,   97,  100,  101,  102,

      235,  235,  235,  101,  235,  235,  235,  235,  235,  235,
      235,  235,  101,  235,  235,  235,  235,  235,  235,  102,
      102,  102,  235,  235,  100,  100,  100,  100,  100,  235,
      235,  235,  101,  235,  101,  101,  101,  101,  101,  101,
      101,  104,  235,  104,  235,  235,  235,  104,  235,  104,
      104,  235,  235,  235,  104,  235,  235,  235,  235,  235,
      235,  235,  104,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  106,  104,  107,  108,  109,  235,  235,
      108,  235,  235,  235,  235,  235,  108,  110,  111,  110,
      108,  108,  108,  110,  108,  110,  110,  109,  109,  109,

      110,  108,  107,  107,  107,  107,  107,  108,  112,  108,
      108,  108,  108,  108,  108,  108,  108,  108,  108,  108,
      110,  108,  104,  235,  104,  235,  235,  235,  104,  235,
      113,  104,  235,  235,  235,  104,  235,  235,  235,  235,
      235,  235,  235,  104,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  106,  104,  114,  115,  116,  235,
      235,  115,  235,  235,  235,  235,  235,  115,  117,  115,
      117,  235,  115,  115,  117,  115,  117,  117,  116,  116,
      116,  117,  115,  114,  114,  114,  114,  114,  115,  118,
      115,  115,  115,  115,  115,  115,  115,  115,  115,  115,

      115,  117,  115,  122,  235,  235,  235,  235,  235,  235,
      123,  124,  124,  124,  235,  235,  235,  125,  235,  126,
      127,  235,  235,  235,  128,  235,  235,  235,  125,  235,
      235,  126,  127,  142,  143,  144,  235,  235,  235,  143,
      235,  235,  235,  235,  235,  235,  235,  235,  143,  235,
      235,  235,  235,  235,  145,  144,  144,  144,  235,  235,
      142,  142,  142,  142,  142,  235,  235,  235,  143,  235,
      143,  143,  143,  143,  143,  143,  143,  147,  148,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  149,
      235,  149,  235,  235,  235,  149,  235,  149,  149,  235,

      235,  235,  150,  235,  147,  147,  147,  147,  147,  235,
      149,  235,  148,  235,  148,  148,  148,  148,  148,  148,
      148,  235,  149,  152,  153,  154,  235,  235,  153,  235,
      235,  235,  235,  235,  155,  156,  157,  156,  155,  153,
      153,  156,  153,  156,  156,  154,  154,  154,  156,  153,
      152,  152,  152,  152,  152,  153,  158,  153,  153,  153,
      159,  153,  153,  153,  159,  153,  153,  153,  156,  153,
      162,  163,  164,  235,  235,  163,  235,  235,  235,  235,
      235,  165,  166,  167,  166,  168,  163,  163,  166,  163,
      166,  166,  164,  164,  164,  166,  163,  162,  162,  162,

      162,  162,  163,  169,  163,  163,  163,  170,  163,  163,
      163,  170,  163,  163,  163,  166,  163,  171,  171,  171,
      171,  171,  171,  171,  171,  172,  171,  171,  173,  104,
      174,  104,  174,  174,  171,  104,  171,  175,  104,  174,
      174,  174,  104,  171,  174,  174,  174,  174,  174,  174,
      104,  171,  174,  171,  174,  174,  174,  174,  174,  174,
      174,  176,  104,  171,  195,  148,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  195,  195,  195,  195,  195,  235,  235,  235,  148,

      235,  148,  148,  148,  148,  148,  148,  148,  196,  197,
      198,  235,  235,  235,  197,  235,  235,  235,  235,  235,
      235,  235,  235,  197,  235,  235,  235,  235,  235,   99,
      198,  198,  198,  235,  235,  196,  196,  196,  196,  196,
      235,  235,  235,  197,  235,  197,  197,  197,  197,  197,
      197,  197,  199,  200,  201,  235,  235,  235,  200,  235,
      235,  235,  235,  235,  235,  235,  235,  200,  235,  235,
      235,  235,  235,  235,  201,  201,  201,  235,  235,  199,
      199,  199,  199,  199,  235,  235,  235,  200,  235,  200,
      200,  200,  200,  200,  200,  200,  218,  219,  220,  235,

      221,  235,  235,  235,  221,  221,  221,  221,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  220,  220,
      220,  235,  235,  218,  218,  218,  218,  218,  235,  235,
      235,  219,  235,  219,  219,  219,  219,  219,  219,  219,
      222,  223,  224,  235,  235,  235,  223,  235,  235,  235,
      235,  235,  235,  235,  235,  223,  235,  235,  235,  235,
      235,  145,  224,  224,  224,  235,  235,  222,  222,  222,
      222,  222,  235,  235,  235,  223,  235,  223,  223,  223,
      223,  223,  223,  223,  232,  233,  234,  235,  229,  235,
      235,  235,  229,  229,  229,  229,  235,  235,  235,  235,

      235,  235,  235,  235,  235,  235,  234,  234,  234,  235,
      235,  232,  232,  232,  232,  232,  235,  235,  235,  233,
      235,  233,  233,  233,  233,  233,  233,  233,   50,   50,
       50,   50,   50,   50,   50,   57,   57,   57,   57,   57,
       57,   57,   63,   63,   63,   63,   63,   63,   63,   70,
       70,   70,   70,   70,   70,   70,   77,   77,   77,   77,
       77,   77,   77,   84,   84,   84,   84,   84,   84,   84,
       28,   28,   28,   28,   28,   28,   28,  105,  105,  105,
      105,  105,  129,  235,  129,  129,  134,  134,  134,  134,
      134,  134,  134,  136,  235,  235,  136,  235,  235,  136,

      137,  235,  137,  137,  137,  235,  137,  191,  191,  191,
      191,  191,  191,  191,  171,  171,  235,  171,  235,  171,
      171,  194,  194,  194,  194,  235,  194,  194,  202,  235,
      202,  202,  203,  235,  203,  203,  205,  235,  205,  205,
      205,  205,  205,   25,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235

    } ;

static const flex_int16_t yy_chk[1892] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    3,   29,   54,
        3,    3,    3,   29,    3,   29,   39,   60,   39,   39,
       39,   41,   41,    3,   46,   46,   46,  123,  123,  123,
      255,   51,  248,   41,   41,   54,   51,   46,   51,   58,
      216,   66,   39,   60,   58,   74,   58,    3,  215,    3,
        4,   64,   81,    4,    4,    4,   64,    4,   64,   88,

       91,   94,  113,   71,  204,  189,    4,   66,   71,   78,
       71,   74,   85,  186,   78,  103,   78,   85,   81,   85,
      103,  185,  103,  193,  138,   88,   91,   94,  113,  138,
        4,  138,    4,    5,  120,  193,    5,    5,    5,  176,
        5,  175,  120,  120,  120,  173,  125,  170,  125,    5,
      125,  125,  125,  126,  126,  120,  169,  128,  128,  128,
      181,  181,  181,  128,  168,  133,  126,  167,  133,  166,
      128,  165,  164,    5,  128,    5,    6,  133,  163,    6,
        6,    6,  162,    6,  130,  130,  130,  130,  131,  131,
      131,  130,    6,  225,  225,  225,  178,  157,  130,  146,

      135,  131,  130,  132,  178,  178,  178,  179,  179,  179,
      127,  117,  182,  182,  182,  116,    6,  178,    6,    7,
      179,  115,    7,    7,    7,  182,    7,  180,  180,  180,
      183,  183,  114,  180,  106,    7,  184,  184,  104,   95,
      180,  213,  213,  183,  180,   93,  188,   89,  188,  184,
      188,  188,  188,   82,  213,  211,  211,  211,   75,    7,
        7,    7,    8,   67,   61,    8,    8,    8,  211,    8,
      208,  208,  208,   55,   48,  209,  208,  209,    8,  209,
      209,  209,   25,  208,  210,  210,  210,  208,    0,    0,
      210,    0,    0,  212,  212,  212,    0,  210,  214,  214,

        0,  210,    8,    8,    8,    9,  212,    0,    9,    9,
        9,  214,    9,  217,  217,  217,    0,    0,  218,    0,
        0,    9,  218,  218,  218,  218,  217,  226,  226,  226,
      227,  227,  227,  228,  228,  228,  230,  230,  230,    0,
      226,    0,    0,  227,    0,    9,  228,    9,   10,  230,
        0,   10,   10,   10,    0,   10,  231,  231,  231,    0,
        0,  232,    0,    0,   10,  232,  232,  232,  232,  231,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,   10,    0,
       10,   11,    0,    0,   11,   11,   11,    0,   11,    0,

        0,    0,    0,    0,    0,    0,    0,   11,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,   11,    0,   11,   12,    0,    0,   12,   12,   12,
        0,   12,    0,    0,    0,    0,    0,    0,    0,    0,
       12,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,   12,    0,   12,   13,    0,    0,
       13,   13,   13,    0,   13,    0,    0,    0,    0,    0,
        0,    0,    0,   13,    0,    0,    0,    0,    0,    0,

        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,   13,    0,   13,
       14,    0,    0,   14,   14,   14,    0,   14,    0,    0,
        0,    0,    0,    0,    0,    0,   14,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
       14,    0,   14,   15,    0,    0,   15,   15,   15,    0,
       15,    0,    0,    0,    0,    0,    0,    0,    0,   15,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,

        0,    0,    0,   15,    0,   15,   16,    0,    0,   16,
       16,   16,    0,   16,    0,    0,    0,    0,    0,    0,
        0,    0,   16,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,   16,    0,   16,   17,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,   17,   17,   17,   17,    0,    0,   17,    0,   17,
       17,   17,   17,   17,   17,    0,   17,   17,   17,   17,
       17,    0,   17,    0,   17,    0,    0,    0,    0,    0,
        0,    0,    0,    0,   17,   19,    0,    0,   19,   19,

       19,    0,   19,    0,    0,    0,    0,    0,    0,    0,
        0,   19,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,   19,    0,   19,   20,    0,
        0,   20,   20,   20,    0,   20,    0,    0,    0,    0,
        0,    0,    0,    0,   20,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,   20,    0,
       20,   21,    0,    0,   21,   21,   21,    0,   21,    0,
        0,    0,    0,    0,    0,    0,    0,   21,    0,    0,

        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,   21,    0,   21,   22,    0,    0,   22,   22,   22,
        0,   22,    0,    0,    0,    0,    0,    0,    0,    0,
       22,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,   22,    0,   22,   23,    0,    0,
       23,   23,   23,    0,   23,    0,    0,    0,    0,    0,
        0,    0,    0,   23,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,

        0,    0,    0,    0,    0,    0,    0,   23,    0,   23,
       24,    0,    0,   24,   24,   24,    0,   24,    0,    0,
        0,    0,    0,    0,    0,    0,   24,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
       24,    0,   24,   26,   26,   26,    0,    0,    0,   26,
        0,    0,    0,    0,    0,    0,    0,    0,   26,    0,
        0,    0,    0,    0,   26,   26,   26,   26,    0,    0,
       26,   26,   26,   26,   26,    0,    0,    0,   26,    0,
       26,   26,   26,   26,   26,   26,   26,   27,   27,   27,

        0,    0,    0,   27,    0,    0,    0,    0,    0,    0,
        0,    0,   27,    0,    0,    0,    0,    0,    0,   27,
       27,   27,    0,    0,   27,   27,   27,   27,   27,    0,
        0,    0,   27,    0,   27,   27,   27,   27,   27,   27,
       27,   32,    0,   32,    0,    0,    0,   32,    0,   32,
       32,    0,    0,    0,   32,    0,    0,    0,    0,    0,
        0,    0,   32,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,   32,   32,   33,   33,   33,    0,    0,
       33,    0,    0,    0,    0,    0,   33,   33,   33,   33,
       33,   33,   33,   33,   33,   33,   33,   33,   33,   33,

       33,   33,   33,   33,   33,   33,   33,   33,   33,   33,
       33,   33,   33,   33,   33,   33,   33,   33,   33,   33,
       33,   33,   34,    0,   34,    0,    0,    0,   34,    0,
       34,   34,    0,    0,    0,   34,    0,    0,    0,    0,
        0,    0,    0,   34,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,   34,   34,   35,   35,   35,    0,
        0,   35,    0,    0,    0,    0,    0,   35,   35,   35,
       35,    0,   35,   35,   35,   35,   35,   35,   35,   35,
       35,   35,   35,   35,   35,   35,   35,   35,   35,   35,
       35,   35,   35,   35,   35,   35,   35,   35,   35,   35,

       35,   35,   35,   40,    0,    0,    0,    0,    0,    0,
       40,   40,   40,   40,    0,    0,    0,   40,    0,   40,
       40,    0,    0,    0,   40,    0,    0,    0,   40,    0,
        0,   40,   40,   92,   92,   92,    0,    0,    0,   92,
        0,    0,    0,    0,    0,    0,    0,    0,   92,    0,
        0,    0,    0,    0,   92,   92,   92,   92,    0,    0,
       92,   92,   92,   92,   92,    0,    0,    0,   92,    0,
       92,   92,   92,   92,   92,   92,   92,   99,   99,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,   99,
        0,   99,    0,    0,    0,   99,    0,   99,   99,    0,

        0,    0,   99,    0,   99,   99,   99,   99,   99,    0,
       99,    0,   99,    0,   99,   99,   99,   99,   99,   99,
       99,    0,   99,  112,  112,  112,    0,    0,  112,    0,
        0,    0,    0,    0,  112,  112,  112,  112,  112,  112,
      112,  112,  112,  112,  112,  112,  112,  112,  112,  112,
      112,  112,  112,  112,  112,  112,  112,  112,  112,  112,
      112,  112,  112,  112,  112,  112,  112,  112,  112,  112,
      118,  118,  118,    0,    0,  118,    0,    0,    0,    0,
        0,  118,  118,  118,  118,  118,  118,  118,  118,  118,
      118,  118,  118,  118,  118,  118,  118,  118,  118,  118,

      118,  118,  118,  118,  118,  118,  118,  118,  118,  118,
      118,  118,  118,  118,  118,  118,  118,  119,  119,  119,
      119,  119,  119,  119,  119,  119,  119,  119,  119,  119,
      119,  119,  119,  119,  119,  119,  119,  119,  119,  119,
      119,  119,  119,  119,  119,  119,  119,  119,  119,  119,
      119,  119,  119,  119,  119,  119,  119,  119,  119,  119,
      119,  119,  119,  119,  145,  145,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,  145,  145,  145,  145,  145,    0,    0,    0,  145,

        0,  145,  145,  145,  145,  145,  145,  145,  147,  147,
      147,    0,    0,    0,  147,    0,    0,    0,    0,    0,
        0,    0,    0,  147,    0,    0,    0,    0,    0,  147,
      147,  147,  147,    0,    0,  147,  147,  147,  147,  147,
        0,    0,    0,  147,    0,  147,  147,  147,  147,  147,
      147,  147,  148,  148,  148,    0,    0,    0,  148,    0,
        0,    0,    0,    0,    0,    0,    0,  148,    0,    0,
        0,    0,    0,    0,  148,  148,  148,    0,    0,  148,
      148,  148,  148,  148,    0,    0,    0,  148,    0,  148,
      148,  148,  148,  148,  148,  148,  192,  192,  192,    0,

      192,    0,    0,    0,  192,  192,  192,  192,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,  192,  192,
      192,    0,    0,  192,  192,  192,  192,  192,    0,    0,
        0,  192,    0,  192,  192,  192,  192,  192,  192,  192,
      195,  195,  195,    0,    0,    0,  195,    0,    0,    0,
        0,    0,    0,    0,    0,  195,    0,    0,    0,    0,
        0,  195,  195,  195,  195,    0,    0,  195,  195,  195,
      195,  195,    0,    0,    0,  195,    0,  195,  195,  195,
      195,  195,  195,  195,  229,  229,  229,    0,  229,    0,
        0,    0,  229,  229,  229,  229,    0,    0,    0,    0,

        0,    0,    0,    0,    0,    0,  229,  229,  229,    0,
        0,  229,  229,  229,  229,  229,    0,    0,    0,  229,
        0,  229,  229,  229,  229,  229,  229,  229,  236,  236,
      236,  236,  236,  236,  236,  237,  237,  237,  237,  237,
      237,  237,  238,  238,  238,  238,  238,  238,  238,  239,
      239,  239,  239,  239,  239,  239,  240,  240,  240,  240,
      240,  240,  240,  241,  241,  241,  241,  241,  241,  241,
      242,  242,  242,  242,  242,  242,  242,  243,  243,  243,
      243,  243,  244,    0,  244,  244,  245,  245,  245,  245,
      245,  245,  245,  246,    0,    0,  246,    0,    0,  246,

      247,    0,  247,  247,  247,    0,  247,  249,  249,  249,
      249,  249,  249,  249,  250,  250,    0,  250,    0,  250,
      250,  251,  251,  251,  251,    0,  251,  251,  252,    0,
      252,  252,  253,    0,  253,  253,  254,    0,  254,  254,
      254,  254,  254,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235,  235,  235,  235,  235,  235,  235,  235,  235,  235,
      235

    } ;

static yy_state_type yy_last_accepting_state;
static char *yy_last_accepting_cpos;

extern int yy_flex_debug;
int yy_flex_debug = 1;

static const flex_int16_t yy_rule_linenum[55] =
    {   0,
      134,  135,  138,  139,  140,  143,  144,  145,  146,  149,
      154,  155,  156,  159,  160,  163,  164,  166,  167,  169,
      170,  174,  176,  190,  192,  194,  198,  199,  200,  201,
      202,  203,  204,  207,  209,  211,  212,  213,  214,  222,
      223,  224,  226,  227,  228,  230,  231,  232,  234,  235,
      238,  241,  242,  244
    } ;

/* The intent behind this definition is that it'll catch
 * any uses of REJECT which flex missed.
 */
#define REJECT reject_used_but_not_detected
#define yymore() yymore_used_but_not_detected
#define YY_MORE_ADJ 0
#define YY_RESTORE_YY_MORE_OFFSET
char *yytext;
#line 1 "lexer.l"
#line 2 "lexer.l"
# include <cerrno>
# include <climits>
# include <cstdlib>
# include <string>
# include <regex>
# include "driver.hh"
# include "parser.hh"

// Work around an incompatibility in flex (at least versions
// 2.5.31 through 2.5.33): it generates code that does
// not conform to C89.  See Debian bug 333231
// <http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=333231>.
# undef yywrap
# define yywrap() 1

// Pacify warnings in yy_init_buffer (observed with Flex 2.6.4)
// and GCC 7.3.0.
#if defined __GNUC__ && 7 <= __GNUC__
# pragma GCC diagnostic ignored "-Wnull-dereference"
#endif

// Forward declarations
yy::parser::symbol_type make_integer(int radix, bool positive, int shift,
				     const yy::parser::location_type& loc);

yy::parser::symbol_type make_boxed_integer(int radix, bool positive, int shift,
				           const yy::parser::location_type& loc);

yy::parser::symbol_type make_integer10(const yy::parser::location_type& loc);

yy::parser::symbol_type make_boxed_integer10(const yy::parser::location_type& loc);

yy::parser::symbol_type make_rational(const yy::parser::location_type& loc);

yy::parser::symbol_type make_char(const yy::parser::location_type& loc, driver& drv);
yy::parser::symbol_type make_string(const yy::parser::location_type& loc, driver& drv);

#line 1071 "lexer.cc"
#define YY_NO_INPUT 1
#define YY_NO_UNISTD_H 1

/* We want to parse incorrect escapes as char/string so that we can give better error messages.
   If we don't do so, then bad escapes fail to lex and fall through to "invalid character" error message. */
#line 120 "lexer.l"
  // Code run each time a pattern is matched.
  # define YY_USER_ACTION  loc.columns (yyleng); drv.step_closing_token();
#line 1080 "lexer.cc"
#line 1081 "lexer.cc"

#define INITIAL 0
#define layout 1
#define layout_do 2
#define layout_if 3
#define layout_left 4
#define bol 5
#define nested_comment 6
#define nested_comment2 7
#define option_prags 8
#define line_prag1a 9
#define line_prag2 10
#define line_prag2a 11

#ifndef YY_NO_UNISTD_H
/* Special case for "unistd.h", since it is non-ANSI. We include it way
 * down here because we want the user's section 1 to have been scanned first.
 * The user has a chance to override it with an option.
 */
/* %if-c-only */
#include <unistd.h>
/* %endif */
/* %if-c++-only */
/* %endif */
#endif

#ifndef YY_EXTRA_TYPE
#define YY_EXTRA_TYPE void *
#endif

/* %if-c-only Reentrant structure and macros (non-C++). */
/* %if-reentrant */
/* %if-c-only */

static int yy_init_globals ( void );

/* %endif */
/* %if-reentrant */
/* %endif */
/* %endif End reentrant structures and macros. */

/* Accessor methods to globals.
   These are made visible to non-reentrant scanners for convenience. */

int yylex_destroy ( void );

int yyget_debug ( void );

void yyset_debug ( int debug_flag  );

YY_EXTRA_TYPE yyget_extra ( void );

void yyset_extra ( YY_EXTRA_TYPE user_defined  );

FILE *yyget_in ( void );

void yyset_in  ( FILE * _in_str  );

FILE *yyget_out ( void );

void yyset_out  ( FILE * _out_str  );

			int yyget_leng ( void );

char *yyget_text ( void );

int yyget_lineno ( void );

void yyset_lineno ( int _line_number  );

/* %if-bison-bridge */
/* %endif */

/* Macros after this point can all be overridden by user definitions in
 * section 1.
 */

#ifndef YY_SKIP_YYWRAP
#ifdef __cplusplus
extern "C" int yywrap ( void );
#else
extern int yywrap ( void );
#endif
#endif

/* %not-for-header */
#ifndef YY_NO_UNPUT
    
#endif
/* %ok-for-header */

/* %endif */

#ifndef yytext_ptr
static void yy_flex_strncpy ( char *, const char *, int );
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen ( const char * );
#endif

#ifndef YY_NO_INPUT
/* %if-c-only Standard (non-C++) definition */
/* %not-for-header */
#ifdef __cplusplus
static int yyinput ( void );
#else
static int input ( void );
#endif
/* %ok-for-header */

/* %endif */
#endif

/* %if-c-only */

        static int yy_start_stack_ptr = 0;
        static int yy_start_stack_depth = 0;
        static int *yy_start_stack = NULL;
    
    static void yy_push_state ( int _new_state );
    
    static void yy_pop_state ( void );
    
    static int yy_top_state ( void );
    
/* %endif */

/* Amount of stuff to slurp up with each read. */
#ifndef YY_READ_BUF_SIZE
#ifdef __ia64__
/* On IA-64, the buffer size is 16k, not 8k */
#define YY_READ_BUF_SIZE 16384
#else
#define YY_READ_BUF_SIZE 8192
#endif /* __ia64__ */
#endif

/* Copy whatever the last rule matched to the standard output. */
#ifndef ECHO
/* %if-c-only Standard (non-C++) definition */
/* This used to be an fputs(), but since the string might contain NUL's,
 * we now use fwrite().
 */
#define ECHO do { if (fwrite( yytext, (size_t) yyleng, 1, yyout )) {} } while (0)
/* %endif */
/* %if-c++-only C++ definition */
/* %endif */
#endif

/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
/* %% [5.0] fread()/read() definition of YY_INPUT goes here unless we're doing C++ \ */\
	if ( YY_CURRENT_BUFFER_LVALUE->yy_is_interactive ) \
		{ \
		int c = '*'; \
		int n; \
		for ( n = 0; n < max_size && \
			     (c = getc( yyin )) != EOF && c != '\n'; ++n ) \
			buf[n] = (char) c; \
		if ( c == '\n' ) \
			buf[n++] = (char) c; \
		if ( c == EOF && ferror( yyin ) ) \
			YY_FATAL_ERROR( "input in flex scanner failed" ); \
		result = n; \
		} \
	else \
		{ \
		errno=0; \
		while ( (result = (int) fread(buf, 1, (yy_size_t) max_size, yyin)) == 0 && ferror(yyin)) \
			{ \
			if( errno != EINTR) \
				{ \
				YY_FATAL_ERROR( "input in flex scanner failed" ); \
				break; \
				} \
			errno=0; \
			clearerr(yyin); \
			} \
		}\
\
/* %if-c++-only C++ definition \ */\
/* %endif */

#endif

/* No semi-colon after return; correct usage is to write "yyterminate();" -
 * we don't want an extra ';' after the "return" because that will cause
 * some compilers to complain about unreachable statements.
 */
#ifndef yyterminate
#define yyterminate() return YY_NULL
#endif

/* Number of entries by which start-condition stack grows. */
#ifndef YY_START_STACK_INCR
#define YY_START_STACK_INCR 25
#endif

/* Report a fatal error. */
#ifndef YY_FATAL_ERROR
/* %if-c-only */
#define YY_FATAL_ERROR(msg) yy_fatal_error( msg )
/* %endif */
/* %if-c++-only */
/* %endif */
#endif

/* %if-tables-serialization structures and prototypes */
/* %not-for-header */
/* %ok-for-header */

/* %not-for-header */
/* %tables-yydmap generated elements */
/* %endif */
/* end tables serialization structures and prototypes */

/* %ok-for-header */

/* Default declaration of generated scanner - a define so the user can
 * easily add parameters.
 */
#ifndef YY_DECL
#define YY_DECL_IS_OURS 1
/* %if-c-only Standard (non-C++) definition */

extern int yylex (void);

#define YY_DECL int yylex (void)
/* %endif */
/* %if-c++-only C++ definition */
/* %endif */
#endif /* !YY_DECL */

/* Code executed at the beginning of each rule, after yytext and yyleng
 * have been set up.
 */
#ifndef YY_USER_ACTION
#define YY_USER_ACTION
#endif

/* Code executed at the end of each rule. */
#ifndef YY_BREAK
#define YY_BREAK /*LINTED*/break;
#endif

/* %% [6.0] YY_RULE_SETUP definition goes here */
#define YY_RULE_SETUP \
	YY_USER_ACTION

/* %not-for-header */
/** The main scanner function which does all the work.
 */
YY_DECL
{
	yy_state_type yy_current_state;
	char *yy_cp, *yy_bp;
	int yy_act;
    
	if ( !(yy_init) )
		{
		(yy_init) = 1;

#ifdef YY_USER_INIT
		YY_USER_INIT;
#endif

		if ( ! (yy_start) )
			(yy_start) = 1;	/* first start state */

		if ( ! yyin )
/* %if-c-only */
			yyin = stdin;
/* %endif */
/* %if-c++-only */
/* %endif */

		if ( ! yyout )
/* %if-c-only */
			yyout = stdout;
/* %endif */
/* %if-c++-only */
/* %endif */

		if ( ! YY_CURRENT_BUFFER ) {
			yyensure_buffer_stack ();
			YY_CURRENT_BUFFER_LVALUE =
				yy_create_buffer( yyin, YY_BUF_SIZE );
		}

		yy_load_buffer_state(  );
		}

	{
/* %% [7.0] user's declarations go here */
#line 124 "lexer.l"


#line 127 "lexer.l"
  /* Code run each time yylex is called. */

  // A handy shortcut to the location held by the driver.
  yy::location& loc = drv.location;
  loc.step ();


#line 1391 "lexer.cc"

	while ( /*CONSTCOND*/1 )		/* loops until end-of-file is reached */
		{
/* %% [8.0] yymore()-related code goes here */
		yy_cp = (yy_c_buf_p);

		/* Support of yytext. */
		*yy_cp = (yy_hold_char);

		/* yy_bp points to the position in yy_ch_buf of the start of
		 * the current run.
		 */
		yy_bp = yy_cp;

/* %% [9.0] code to set up and find next match goes here */
		yy_current_state = (yy_start);
yy_match:
		do
			{
			YY_CHAR yy_c = yy_ec[YY_SC_TO_UI(*yy_cp)] ;
			if ( yy_accept[yy_current_state] )
				{
				(yy_last_accepting_state) = yy_current_state;
				(yy_last_accepting_cpos) = yy_cp;
				}
			while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
				{
				yy_current_state = (int) yy_def[yy_current_state];
				if ( yy_current_state >= 236 )
					yy_c = yy_meta[yy_c];
				}
			yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
			++yy_cp;
			}
		while ( yy_current_state != 235 );
		yy_cp = (yy_last_accepting_cpos);
		yy_current_state = (yy_last_accepting_state);

yy_find_action:
/* %% [10.0] code to find the action number goes here */
		yy_act = yy_accept[yy_current_state];

		YY_DO_BEFORE_ACTION;

/* %% [11.0] code for yylineno update goes here */

do_action:	/* This label is used only to access EOF actions. */

/* %% [12.0] debug code goes here */
		if ( yy_flex_debug )
			{
			if ( yy_act == 0 )
				fprintf( stderr, "--scanner backing up\n" );
			else if ( yy_act < 55 )
				fprintf( stderr, "--accepting rule at line %ld (\"%s\")\n",
				         (long)yy_rule_linenum[yy_act], yytext );
			else if ( yy_act == 55 )
				fprintf( stderr, "--accepting default rule (\"%s\")\n",
				         yytext );
			else if ( yy_act == 56 )
				fprintf( stderr, "--(end of buffer or a NUL)\n" );
			else
				fprintf( stderr, "--EOF (start condition %d)\n", YY_START );
			}

		switch ( yy_act )
	{ /* beginning of action switch */
/* %% [13.0] actions go here */
			case 0: /* must back up */
			/* undo the effects of YY_DO_BEFORE_ACTION */
			*yy_cp = (yy_hold_char);
			yy_cp = (yy_last_accepting_cpos);
			yy_current_state = (yy_last_accepting_state);
			goto yy_find_action;

case 1:
YY_RULE_SETUP
#line 134 "lexer.l"
loc.step ();
	YY_BREAK
case 2:
YY_RULE_SETUP
#line 135 "lexer.l"
loc.step ();
	YY_BREAK
/* GHC calls some handwritten code to deal with nested comments */
case 3:
YY_RULE_SETUP
#line 138 "lexer.l"
yy_pop_state();
	YY_BREAK
case 4:
YY_RULE_SETUP
#line 139 "lexer.l"
loc.step(); yy_pop_state();
	YY_BREAK
case 5:
YY_RULE_SETUP
#line 140 "lexer.l"
throw yy::parser::syntax_error(loc, "Too many -}");
	YY_BREAK

case 6:
/* rule 6 can match eol */
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
YY_LINENO_REWIND_TO(yy_bp + 2);
(yy_c_buf_p) = yy_cp = yy_bp + 2;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 143 "lexer.l"
yy_push_state(nested_comment2);
	YY_BREAK
case 7:
YY_RULE_SETUP
#line 144 "lexer.l"
// eat a chunk with no newline or "-"
	YY_BREAK
case 8:
YY_RULE_SETUP
#line 145 "lexer.l"
// eat a "-" not followed by "}"
	YY_BREAK
case 9:
/* rule 9 can match eol */
YY_RULE_SETUP
#line 146 "lexer.l"
loc.lines(1);
	YY_BREAK
case YY_STATE_EOF(nested_comment):
case YY_STATE_EOF(nested_comment2):
#line 147 "lexer.l"
throw yy::parser::syntax_error(loc, "Unterminated {-");
	YY_BREAK

case 10:
/* rule 10 can match eol */
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
YY_LINENO_REWIND_TO(yy_bp + 2);
(yy_c_buf_p) = yy_cp = yy_bp + 2;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 149 "lexer.l"
yy_push_state(nested_comment);
	YY_BREAK
case 11:
YY_RULE_SETUP
#line 154 "lexer.l"
loc.step();
	YY_BREAK
case 12:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 155 "lexer.l"
loc.step();
	YY_BREAK
case 13:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 156 "lexer.l"
loc.step();
	YY_BREAK
/* bol state */
case 14:
/* rule 14 can match eol */
YY_RULE_SETUP
#line 159 "lexer.l"
loc.lines(1); loc.step();
	YY_BREAK
case 15:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 160 "lexer.l"
if (auto tok = drv.do_bol(loc)) return *tok;
	YY_BREAK
case YY_STATE_EOF(bol):
#line 161 "lexer.l"
if (auto tok = drv.do_bol(loc)) return *tok;
	YY_BREAK
case 16:
/* rule 16 can match eol */
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
YY_LINENO_REWIND_TO(yy_bp + 1);
(yy_c_buf_p) = yy_cp = yy_bp + 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 163 "lexer.l"
return drv.hopefully_open_brace(loc);
	YY_BREAK
case 17:
/* rule 17 can match eol */
YY_RULE_SETUP
#line 164 "lexer.l"
loc.lines(1); loc.step();
	YY_BREAK
case 18:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp = yy_bp + 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 166 "lexer.l"
drv.hopefully_open_brace(loc);
	YY_BREAK
case 19:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 167 "lexer.l"
yy_pop_state();
	YY_BREAK
case 20:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 169 "lexer.l"
return drv.new_layout_context(loc, true, true, yy::parser::token::TOK_VOCURLY);
	YY_BREAK
case 21:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 170 "lexer.l"
return drv.new_layout_context(loc, false, true, yy::parser::token::TOK_VOCURLY);
	YY_BREAK
/* If we've generated a VOCURLY and started a new line to the left, then
     we need to generate a VCCURLY */
case 22:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 174 "lexer.l"
return drv.do_layout_left(loc);
	YY_BREAK
case 23:
/* rule 23 can match eol */
YY_RULE_SETUP
#line 176 "lexer.l"
loc.lines(1); loc.step (); yy_push_state(bol);
	YY_BREAK
/* line pragmas */
/* <line_prag1> single-line line pragmas: # <line> "<file>" <extra-stuff> \n */
/* <line_prag2> Haskell-style line pragms, of the form
     {-# LINE <line> "<file>" #-} */
/* <line_prag2a> */
/* <column_prag> */
/* two-word pragmas */
case 24:
/* rule 24 can match eol */
YY_RULE_SETUP
#line 190 "lexer.l"
if (auto prag = drv.prag(loc)) return *prag; else yy_push_state(nested_comment);
	YY_BREAK
/* one-word pragmas */
case 25:
/* rule 25 can match eol */
YY_RULE_SETUP
#line 192 "lexer.l"
if (auto prag = drv.prag(loc)) return *prag; else yy_push_state(nested_comment);
	YY_BREAK
/* end pragma */
case 26:
YY_RULE_SETUP
#line 194 "lexer.l"
return yy::parser::make_CLOSE_PRAG(loc);
	YY_BREAK
/* unboxed tuples: (# and #) */
case 27:
YY_RULE_SETUP
#line 198 "lexer.l"
return yy::parser::make_OPAREN (loc);
	YY_BREAK
case 28:
YY_RULE_SETUP
#line 199 "lexer.l"
drv.set_closing_token(); return yy::parser::make_CPAREN (loc);
	YY_BREAK
case 29:
YY_RULE_SETUP
#line 200 "lexer.l"
return yy::parser::make_OBRACK (loc);
	YY_BREAK
case 30:
YY_RULE_SETUP
#line 201 "lexer.l"
drv.set_closing_token(); return yy::parser::make_CBRACK (loc);
	YY_BREAK
case 31:
YY_RULE_SETUP
#line 202 "lexer.l"
return yy::parser::make_COMMA  (loc);
	YY_BREAK
case 32:
YY_RULE_SETUP
#line 203 "lexer.l"
return yy::parser::make_SEMI   (loc);
	YY_BREAK
case 33:
YY_RULE_SETUP
#line 204 "lexer.l"
return yy::parser::make_BACKQUOTE (loc);
	YY_BREAK
/* action: push a "NoLayout" context into the stack */
case 34:
YY_RULE_SETUP
#line 207 "lexer.l"
drv.push_context(); return yy::parser::make_OCURLY (loc);
	YY_BREAK
/* action: pop the top context from the stack */
case 35:
YY_RULE_SETUP
#line 209 "lexer.l"
drv.set_closing_token(); drv.pop_context(); return yy::parser::make_CCURLY (loc);
	YY_BREAK
case 36:
YY_RULE_SETUP
#line 211 "lexer.l"
drv.set_closing_token(); return yy::parser::make_QVARID (yytext, loc);
	YY_BREAK
case 37:
YY_RULE_SETUP
#line 212 "lexer.l"
drv.set_closing_token(); return yy::parser::make_QCONID (yytext, loc);
	YY_BREAK
case 38:
YY_RULE_SETUP
#line 213 "lexer.l"
drv.set_closing_token(); return drv.varid(loc);
	YY_BREAK
case 39:
YY_RULE_SETUP
#line 214 "lexer.l"
drv.set_closing_token(); return yy::parser::make_CONID  (yytext, loc);
	YY_BREAK
/* Here we look for {qvarid}#+ ... {conid}#+ if magicHashEnabled */
/* precededByClosing and followedByOpening -> tight infix
     followedByOpening                       -> prefix
     precededByClosing                       -> suffix
     otherwise                               -> loose infix */
case 40:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 2;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 222 "lexer.l"
return drv.varsym(drv.check_closing_token(), false, loc);
	YY_BREAK
case 41:
*yy_cp = (yy_hold_char); /* undo effects of setting up yytext */
(yy_c_buf_p) = yy_cp -= 1;
YY_DO_BEFORE_ACTION; /* set up yytext again */
YY_RULE_SETUP
#line 223 "lexer.l"
return drv.varsym(drv.check_closing_token(), true, loc);
	YY_BREAK
case 42:
YY_RULE_SETUP
#line 224 "lexer.l"
return drv.varsym(drv.check_closing_token(), false, loc);
	YY_BREAK
case 43:
YY_RULE_SETUP
#line 226 "lexer.l"
return yy::parser::make_QVARSYM  (yytext,loc);
	YY_BREAK
case 44:
YY_RULE_SETUP
#line 227 "lexer.l"
return yy::parser::make_QCONSYM  (yytext,loc);
	YY_BREAK
case 45:
YY_RULE_SETUP
#line 228 "lexer.l"
return drv.consym(loc);
	YY_BREAK
case 46:
YY_RULE_SETUP
#line 230 "lexer.l"
drv.set_closing_token(); return make_boxed_integer10(loc);
	YY_BREAK
case 47:
YY_RULE_SETUP
#line 231 "lexer.l"
drv.set_closing_token(); return make_boxed_integer10(loc);
	YY_BREAK
case 48:
YY_RULE_SETUP
#line 232 "lexer.l"
drv.set_closing_token(); return make_integer10(loc);
	YY_BREAK
/* 0[bB]{numspc}{binary}      make_integer(2,true,2,loc); */
case 49:
YY_RULE_SETUP
#line 234 "lexer.l"
drv.set_closing_token(); return make_integer( 8,true,2,loc);
	YY_BREAK
case 50:
YY_RULE_SETUP
#line 235 "lexer.l"
drv.set_closing_token(); return make_integer(16,true,2,loc);
	YY_BREAK
/* negative literals depend on an extension */
case 51:
YY_RULE_SETUP
#line 238 "lexer.l"
drv.set_closing_token(); return make_rational(loc);
	YY_BREAK
/* Its important that we only allow escaped quotes inside char or string literals */
case 52:
YY_RULE_SETUP
#line 241 "lexer.l"
drv.set_closing_token(); return make_char(loc, drv);
	YY_BREAK
case 53:
YY_RULE_SETUP
#line 242 "lexer.l"
drv.set_closing_token(); return make_string(loc, drv);
	YY_BREAK
case 54:
YY_RULE_SETUP
#line 244 "lexer.l"
{
             throw yy::parser::syntax_error
               (loc, "invalid character: " + std::string(yytext));
}
	YY_BREAK
case YY_STATE_EOF(INITIAL):
case YY_STATE_EOF(layout):
case YY_STATE_EOF(layout_do):
case YY_STATE_EOF(layout_if):
case YY_STATE_EOF(layout_left):
case YY_STATE_EOF(option_prags):
case YY_STATE_EOF(line_prag1a):
case YY_STATE_EOF(line_prag2):
case YY_STATE_EOF(line_prag2a):
#line 249 "lexer.l"
return yy::parser::make_END (loc);
	YY_BREAK
case 55:
YY_RULE_SETUP
#line 250 "lexer.l"
ECHO;
	YY_BREAK
#line 1843 "lexer.cc"

	case YY_END_OF_BUFFER:
		{
		/* Amount of text matched not including the EOB char. */
		int yy_amount_of_matched_text = (int) (yy_cp - (yytext_ptr)) - 1;

		/* Undo the effects of YY_DO_BEFORE_ACTION. */
		*yy_cp = (yy_hold_char);
		YY_RESTORE_YY_MORE_OFFSET

		if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_NEW )
			{
			/* We're scanning a new file or input source.  It's
			 * possible that this happened because the user
			 * just pointed yyin at a new source and called
			 * yylex().  If so, then we have to assure
			 * consistency between YY_CURRENT_BUFFER and our
			 * globals.  Here is the right place to do so, because
			 * this is the first action (other than possibly a
			 * back-up) that will match for the new input source.
			 */
			(yy_n_chars) = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
/* %if-c-only */
			YY_CURRENT_BUFFER_LVALUE->yy_input_file = yyin;
/* %endif */
/* %if-c++-only */
/* %endif */
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status = YY_BUFFER_NORMAL;
			}

		/* Note that here we test for yy_c_buf_p "<=" to the position
		 * of the first EOB in the buffer, since yy_c_buf_p will
		 * already have been incremented past the NUL character
		 * (since all states make transitions on EOB to the
		 * end-of-buffer state).  Contrast this with the test
		 * in input().
		 */
		if ( (yy_c_buf_p) <= &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars)] )
			{ /* This was really a NUL. */
			yy_state_type yy_next_state;

			(yy_c_buf_p) = (yytext_ptr) + yy_amount_of_matched_text;

			yy_current_state = yy_get_previous_state(  );

			/* Okay, we're now positioned to make the NUL
			 * transition.  We couldn't have
			 * yy_get_previous_state() go ahead and do it
			 * for us because it doesn't know how to deal
			 * with the possibility of jamming (and we don't
			 * want to build jamming into it because then it
			 * will run more slowly).
			 */

			yy_next_state = yy_try_NUL_trans( yy_current_state );

			yy_bp = (yytext_ptr) + YY_MORE_ADJ;

			if ( yy_next_state )
				{
				/* Consume the NUL. */
				yy_cp = ++(yy_c_buf_p);
				yy_current_state = yy_next_state;
				goto yy_match;
				}

			else
				{
/* %% [14.0] code to do back-up for compressed tables and set up yy_cp goes here */
				yy_cp = (yy_last_accepting_cpos);
				yy_current_state = (yy_last_accepting_state);
				goto yy_find_action;
				}
			}

		else switch ( yy_get_next_buffer(  ) )
			{
			case EOB_ACT_END_OF_FILE:
				{
				(yy_did_buffer_switch_on_eof) = 0;

				if ( yywrap(  ) )
					{
					/* Note: because we've taken care in
					 * yy_get_next_buffer() to have set up
					 * yytext, we can now set up
					 * yy_c_buf_p so that if some total
					 * hoser (like flex itself) wants to
					 * call the scanner after we return the
					 * YY_NULL, it'll still work - another
					 * YY_NULL will get returned.
					 */
					(yy_c_buf_p) = (yytext_ptr) + YY_MORE_ADJ;

					yy_act = YY_STATE_EOF(YY_START);
					goto do_action;
					}

				else
					{
					if ( ! (yy_did_buffer_switch_on_eof) )
						YY_NEW_FILE;
					}
				break;
				}

			case EOB_ACT_CONTINUE_SCAN:
				(yy_c_buf_p) =
					(yytext_ptr) + yy_amount_of_matched_text;

				yy_current_state = yy_get_previous_state(  );

				yy_cp = (yy_c_buf_p);
				yy_bp = (yytext_ptr) + YY_MORE_ADJ;
				goto yy_match;

			case EOB_ACT_LAST_MATCH:
				(yy_c_buf_p) =
				&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars)];

				yy_current_state = yy_get_previous_state(  );

				yy_cp = (yy_c_buf_p);
				yy_bp = (yytext_ptr) + YY_MORE_ADJ;
				goto yy_find_action;
			}
		break;
		}

	default:
		YY_FATAL_ERROR(
			"fatal flex scanner internal error--no action found" );
	} /* end of action switch */
		} /* end of scanning one token */
	} /* end of user's declarations */
} /* end of yylex */
/* %ok-for-header */

/* %if-c++-only */
/* %not-for-header */
/* %ok-for-header */

/* %endif */

/* yy_get_next_buffer - try to read in a new buffer
 *
 * Returns a code representing an action:
 *	EOB_ACT_LAST_MATCH -
 *	EOB_ACT_CONTINUE_SCAN - continue scanning from current position
 *	EOB_ACT_END_OF_FILE - end of file
 */
/* %if-c-only */
static int yy_get_next_buffer (void)
/* %endif */
/* %if-c++-only */
/* %endif */
{
    	char *dest = YY_CURRENT_BUFFER_LVALUE->yy_ch_buf;
	char *source = (yytext_ptr);
	int number_to_move, i;
	int ret_val;

	if ( (yy_c_buf_p) > &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars) + 1] )
		YY_FATAL_ERROR(
		"fatal flex scanner internal error--end of buffer missed" );

	if ( YY_CURRENT_BUFFER_LVALUE->yy_fill_buffer == 0 )
		{ /* Don't try to fill the buffer, so this is an EOF. */
		if ( (yy_c_buf_p) - (yytext_ptr) - YY_MORE_ADJ == 1 )
			{
			/* We matched a single character, the EOB, so
			 * treat this as a final EOF.
			 */
			return EOB_ACT_END_OF_FILE;
			}

		else
			{
			/* We matched some text prior to the EOB, first
			 * process it.
			 */
			return EOB_ACT_LAST_MATCH;
			}
		}

	/* Try to read more data. */

	/* First move last chars to start of buffer. */
	number_to_move = (int) ((yy_c_buf_p) - (yytext_ptr) - 1);

	for ( i = 0; i < number_to_move; ++i )
		*(dest++) = *(source++);

	if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_EOF_PENDING )
		/* don't do the read, it's not guaranteed to return an EOF,
		 * just force an EOF
		 */
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = (yy_n_chars) = 0;

	else
		{
			int num_to_read =
			YY_CURRENT_BUFFER_LVALUE->yy_buf_size - number_to_move - 1;

		while ( num_to_read <= 0 )
			{ /* Not enough room in the buffer - grow it. */

			/* just a shorter name for the current buffer */
			YY_BUFFER_STATE b = YY_CURRENT_BUFFER_LVALUE;

			int yy_c_buf_p_offset =
				(int) ((yy_c_buf_p) - b->yy_ch_buf);

			if ( b->yy_is_our_buffer )
				{
				int new_size = b->yy_buf_size * 2;

				if ( new_size <= 0 )
					b->yy_buf_size += b->yy_buf_size / 8;
				else
					b->yy_buf_size *= 2;

				b->yy_ch_buf = (char *)
					/* Include room in for 2 EOB chars. */
					yyrealloc( (void *) b->yy_ch_buf,
							 (yy_size_t) (b->yy_buf_size + 2)  );
				}
			else
				/* Can't grow it, we don't own it. */
				b->yy_ch_buf = NULL;

			if ( ! b->yy_ch_buf )
				YY_FATAL_ERROR(
				"fatal error - scanner input buffer overflow" );

			(yy_c_buf_p) = &b->yy_ch_buf[yy_c_buf_p_offset];

			num_to_read = YY_CURRENT_BUFFER_LVALUE->yy_buf_size -
						number_to_move - 1;

			}

		if ( num_to_read > YY_READ_BUF_SIZE )
			num_to_read = YY_READ_BUF_SIZE;

		/* Read in more data. */
		YY_INPUT( (&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move]),
			(yy_n_chars), num_to_read );

		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = (yy_n_chars);
		}

	if ( (yy_n_chars) == 0 )
		{
		if ( number_to_move == YY_MORE_ADJ )
			{
			ret_val = EOB_ACT_END_OF_FILE;
			yyrestart( yyin  );
			}

		else
			{
			ret_val = EOB_ACT_LAST_MATCH;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status =
				YY_BUFFER_EOF_PENDING;
			}
		}

	else
		ret_val = EOB_ACT_CONTINUE_SCAN;

	if (((yy_n_chars) + number_to_move) > YY_CURRENT_BUFFER_LVALUE->yy_buf_size) {
		/* Extend the array by 50%, plus the number we really need. */
		int new_size = (yy_n_chars) + number_to_move + ((yy_n_chars) >> 1);
		YY_CURRENT_BUFFER_LVALUE->yy_ch_buf = (char *) yyrealloc(
			(void *) YY_CURRENT_BUFFER_LVALUE->yy_ch_buf, (yy_size_t) new_size  );
		if ( ! YY_CURRENT_BUFFER_LVALUE->yy_ch_buf )
			YY_FATAL_ERROR( "out of dynamic memory in yy_get_next_buffer()" );
		/* "- 2" to take care of EOB's */
		YY_CURRENT_BUFFER_LVALUE->yy_buf_size = (int) (new_size - 2);
	}

	(yy_n_chars) += number_to_move;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars)] = YY_END_OF_BUFFER_CHAR;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars) + 1] = YY_END_OF_BUFFER_CHAR;

	(yytext_ptr) = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[0];

	return ret_val;
}

/* yy_get_previous_state - get the state just before the EOB char was reached */

/* %if-c-only */
/* %not-for-header */
    static yy_state_type yy_get_previous_state (void)
/* %endif */
/* %if-c++-only */
/* %endif */
{
	yy_state_type yy_current_state;
	char *yy_cp;
    
/* %% [15.0] code to get the start state into yy_current_state goes here */
	yy_current_state = (yy_start);

	for ( yy_cp = (yytext_ptr) + YY_MORE_ADJ; yy_cp < (yy_c_buf_p); ++yy_cp )
		{
/* %% [16.0] code to find the next state goes here */
		YY_CHAR yy_c = (*yy_cp ? yy_ec[YY_SC_TO_UI(*yy_cp)] : 4);
		if ( yy_accept[yy_current_state] )
			{
			(yy_last_accepting_state) = yy_current_state;
			(yy_last_accepting_cpos) = yy_cp;
			}
		while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
			{
			yy_current_state = (int) yy_def[yy_current_state];
			if ( yy_current_state >= 236 )
				yy_c = yy_meta[yy_c];
			}
		yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
		}

	return yy_current_state;
}

/* yy_try_NUL_trans - try to make a transition on the NUL character
 *
 * synopsis
 *	next_state = yy_try_NUL_trans( current_state );
 */
/* %if-c-only */
    static yy_state_type yy_try_NUL_trans  (yy_state_type yy_current_state )
/* %endif */
/* %if-c++-only */
/* %endif */
{
	int yy_is_jam;
    /* %% [17.0] code to find the next state, and perhaps do backing up, goes here */
	char *yy_cp = (yy_c_buf_p);

	YY_CHAR yy_c = 4;
	if ( yy_accept[yy_current_state] )
		{
		(yy_last_accepting_state) = yy_current_state;
		(yy_last_accepting_cpos) = yy_cp;
		}
	while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
		{
		yy_current_state = (int) yy_def[yy_current_state];
		if ( yy_current_state >= 236 )
			yy_c = yy_meta[yy_c];
		}
	yy_current_state = yy_nxt[yy_base[yy_current_state] + yy_c];
	yy_is_jam = (yy_current_state == 235);

		return yy_is_jam ? 0 : yy_current_state;
}

#ifndef YY_NO_UNPUT
/* %if-c-only */

/* %endif */
#endif

/* %if-c-only */
#ifndef YY_NO_INPUT
#ifdef __cplusplus
    static int yyinput (void)
#else
    static int input  (void)
#endif

/* %endif */
/* %if-c++-only */
/* %endif */
{
	int c;
    
	*(yy_c_buf_p) = (yy_hold_char);

	if ( *(yy_c_buf_p) == YY_END_OF_BUFFER_CHAR )
		{
		/* yy_c_buf_p now points to the character we want to return.
		 * If this occurs *before* the EOB characters, then it's a
		 * valid NUL; if not, then we've hit the end of the buffer.
		 */
		if ( (yy_c_buf_p) < &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[(yy_n_chars)] )
			/* This was really a NUL. */
			*(yy_c_buf_p) = '\0';

		else
			{ /* need more input */
			int offset = (int) ((yy_c_buf_p) - (yytext_ptr));
			++(yy_c_buf_p);

			switch ( yy_get_next_buffer(  ) )
				{
				case EOB_ACT_LAST_MATCH:
					/* This happens because yy_g_n_b()
					 * sees that we've accumulated a
					 * token and flags that we need to
					 * try matching the token before
					 * proceeding.  But for input(),
					 * there's no matching to consider.
					 * So convert the EOB_ACT_LAST_MATCH
					 * to EOB_ACT_END_OF_FILE.
					 */

					/* Reset buffer status. */
					yyrestart( yyin );

					/*FALLTHROUGH*/

				case EOB_ACT_END_OF_FILE:
					{
					if ( yywrap(  ) )
						return 0;

					if ( ! (yy_did_buffer_switch_on_eof) )
						YY_NEW_FILE;
#ifdef __cplusplus
					return yyinput();
#else
					return input();
#endif
					}

				case EOB_ACT_CONTINUE_SCAN:
					(yy_c_buf_p) = (yytext_ptr) + offset;
					break;
				}
			}
		}

	c = *(unsigned char *) (yy_c_buf_p);	/* cast for 8-bit char's */
	*(yy_c_buf_p) = '\0';	/* preserve yytext */
	(yy_hold_char) = *++(yy_c_buf_p);

/* %% [19.0] update BOL and yylineno */

	return c;
}
/* %if-c-only */
#endif	/* ifndef YY_NO_INPUT */
/* %endif */

/** Immediately switch to a different input stream.
 * @param input_file A readable stream.
 * 
 * @note This function does not reset the start condition to @c INITIAL .
 */
/* %if-c-only */
    void yyrestart  (FILE * input_file )
/* %endif */
/* %if-c++-only */
/* %endif */
{
    
	if ( ! YY_CURRENT_BUFFER ){
        yyensure_buffer_stack ();
		YY_CURRENT_BUFFER_LVALUE =
            yy_create_buffer( yyin, YY_BUF_SIZE );
	}

	yy_init_buffer( YY_CURRENT_BUFFER, input_file );
	yy_load_buffer_state(  );
}

/* %if-c++-only */
/* %endif */

/** Switch to a different input buffer.
 * @param new_buffer The new input buffer.
 * 
 */
/* %if-c-only */
    void yy_switch_to_buffer  (YY_BUFFER_STATE  new_buffer )
/* %endif */
/* %if-c++-only */
/* %endif */
{
    
	/* TODO. We should be able to replace this entire function body
	 * with
	 *		yypop_buffer_state();
	 *		yypush_buffer_state(new_buffer);
     */
	yyensure_buffer_stack ();
	if ( YY_CURRENT_BUFFER == new_buffer )
		return;

	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*(yy_c_buf_p) = (yy_hold_char);
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = (yy_c_buf_p);
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = (yy_n_chars);
		}

	YY_CURRENT_BUFFER_LVALUE = new_buffer;
	yy_load_buffer_state(  );

	/* We don't actually know whether we did this switch during
	 * EOF (yywrap()) processing, but the only time this flag
	 * is looked at is after yywrap() is called, so it's safe
	 * to go ahead and always set it.
	 */
	(yy_did_buffer_switch_on_eof) = 1;
}

/* %if-c-only */
static void yy_load_buffer_state  (void)
/* %endif */
/* %if-c++-only */
/* %endif */
{
    	(yy_n_chars) = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
	(yytext_ptr) = (yy_c_buf_p) = YY_CURRENT_BUFFER_LVALUE->yy_buf_pos;
/* %if-c-only */
	yyin = YY_CURRENT_BUFFER_LVALUE->yy_input_file;
/* %endif */
/* %if-c++-only */
/* %endif */
	(yy_hold_char) = *(yy_c_buf_p);
}

/** Allocate and initialize an input buffer state.
 * @param file A readable stream.
 * @param size The character buffer size in bytes. When in doubt, use @c YY_BUF_SIZE.
 * 
 * @return the allocated buffer state.
 */
/* %if-c-only */
    YY_BUFFER_STATE yy_create_buffer  (FILE * file, int  size )
/* %endif */
/* %if-c++-only */
/* %endif */
{
	YY_BUFFER_STATE b;
    
	b = (YY_BUFFER_STATE) yyalloc( sizeof( struct yy_buffer_state )  );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in yy_create_buffer()" );

	b->yy_buf_size = size;

	/* yy_ch_buf has to be 2 characters longer than the size given because
	 * we need to put in 2 end-of-buffer characters.
	 */
	b->yy_ch_buf = (char *) yyalloc( (yy_size_t) (b->yy_buf_size + 2)  );
	if ( ! b->yy_ch_buf )
		YY_FATAL_ERROR( "out of dynamic memory in yy_create_buffer()" );

	b->yy_is_our_buffer = 1;

	yy_init_buffer( b, file );

	return b;
}

/* %if-c++-only */
/* %endif */

/** Destroy the buffer.
 * @param b a buffer created with yy_create_buffer()
 * 
 */
/* %if-c-only */
    void yy_delete_buffer (YY_BUFFER_STATE  b )
/* %endif */
/* %if-c++-only */
/* %endif */
{
    
	if ( ! b )
		return;

	if ( b == YY_CURRENT_BUFFER ) /* Not sure if we should pop here. */
		YY_CURRENT_BUFFER_LVALUE = (YY_BUFFER_STATE) 0;

	if ( b->yy_is_our_buffer )
		yyfree( (void *) b->yy_ch_buf  );

	yyfree( (void *) b  );
}

/* Initializes or reinitializes a buffer.
 * This function is sometimes called more than once on the same buffer,
 * such as during a yyrestart() or at EOF.
 */
/* %if-c-only */
    static void yy_init_buffer  (YY_BUFFER_STATE  b, FILE * file )
/* %endif */
/* %if-c++-only */
/* %endif */

{
	int oerrno = errno;
    
	yy_flush_buffer( b );

/* %if-c-only */
	b->yy_input_file = file;
/* %endif */
/* %if-c++-only */
/* %endif */
	b->yy_fill_buffer = 1;

    /* If b is the current buffer, then yy_init_buffer was _probably_
     * called from yyrestart() or through yy_get_next_buffer.
     * In that case, we don't want to reset the lineno or column.
     */
    if (b != YY_CURRENT_BUFFER){
        b->yy_bs_lineno = 1;
        b->yy_bs_column = 0;
    }

/* %if-c-only */

        b->yy_is_interactive = 0;
    
/* %endif */
/* %if-c++-only */
/* %endif */
	errno = oerrno;
}

/** Discard all buffered characters. On the next scan, YY_INPUT will be called.
 * @param b the buffer state to be flushed, usually @c YY_CURRENT_BUFFER.
 * 
 */
/* %if-c-only */
    void yy_flush_buffer (YY_BUFFER_STATE  b )
/* %endif */
/* %if-c++-only */
/* %endif */
{
    	if ( ! b )
		return;

	b->yy_n_chars = 0;

	/* We always need two end-of-buffer characters.  The first causes
	 * a transition to the end-of-buffer state.  The second causes
	 * a jam in that state.
	 */
	b->yy_ch_buf[0] = YY_END_OF_BUFFER_CHAR;
	b->yy_ch_buf[1] = YY_END_OF_BUFFER_CHAR;

	b->yy_buf_pos = &b->yy_ch_buf[0];

	b->yy_at_bol = 1;
	b->yy_buffer_status = YY_BUFFER_NEW;

	if ( b == YY_CURRENT_BUFFER )
		yy_load_buffer_state(  );
}

/* %if-c-or-c++ */
/** Pushes the new state onto the stack. The new state becomes
 *  the current state. This function will allocate the stack
 *  if necessary.
 *  @param new_buffer The new state.
 *  
 */
/* %if-c-only */
void yypush_buffer_state (YY_BUFFER_STATE new_buffer )
/* %endif */
/* %if-c++-only */
/* %endif */
{
    	if (new_buffer == NULL)
		return;

	yyensure_buffer_stack();

	/* This block is copied from yy_switch_to_buffer. */
	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*(yy_c_buf_p) = (yy_hold_char);
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = (yy_c_buf_p);
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = (yy_n_chars);
		}

	/* Only push if top exists. Otherwise, replace top. */
	if (YY_CURRENT_BUFFER)
		(yy_buffer_stack_top)++;
	YY_CURRENT_BUFFER_LVALUE = new_buffer;

	/* copied from yy_switch_to_buffer. */
	yy_load_buffer_state(  );
	(yy_did_buffer_switch_on_eof) = 1;
}
/* %endif */

/* %if-c-or-c++ */
/** Removes and deletes the top of the stack, if present.
 *  The next element becomes the new top.
 *  
 */
/* %if-c-only */
void yypop_buffer_state (void)
/* %endif */
/* %if-c++-only */
/* %endif */
{
    	if (!YY_CURRENT_BUFFER)
		return;

	yy_delete_buffer(YY_CURRENT_BUFFER );
	YY_CURRENT_BUFFER_LVALUE = NULL;
	if ((yy_buffer_stack_top) > 0)
		--(yy_buffer_stack_top);

	if (YY_CURRENT_BUFFER) {
		yy_load_buffer_state(  );
		(yy_did_buffer_switch_on_eof) = 1;
	}
}
/* %endif */

/* %if-c-or-c++ */
/* Allocates the stack if it does not exist.
 *  Guarantees space for at least one push.
 */
/* %if-c-only */
static void yyensure_buffer_stack (void)
/* %endif */
/* %if-c++-only */
/* %endif */
{
	yy_size_t num_to_alloc;
    
	if (!(yy_buffer_stack)) {

		/* First allocation is just for 2 elements, since we don't know if this
		 * scanner will even need a stack. We use 2 instead of 1 to avoid an
		 * immediate realloc on the next call.
         */
      num_to_alloc = 1; /* After all that talk, this was set to 1 anyways... */
		(yy_buffer_stack) = (struct yy_buffer_state**)yyalloc
								(num_to_alloc * sizeof(struct yy_buffer_state*)
								);
		if ( ! (yy_buffer_stack) )
			YY_FATAL_ERROR( "out of dynamic memory in yyensure_buffer_stack()" );

		memset((yy_buffer_stack), 0, num_to_alloc * sizeof(struct yy_buffer_state*));

		(yy_buffer_stack_max) = num_to_alloc;
		(yy_buffer_stack_top) = 0;
		return;
	}

	if ((yy_buffer_stack_top) >= ((yy_buffer_stack_max)) - 1){

		/* Increase the buffer to prepare for a possible push. */
		yy_size_t grow_size = 8 /* arbitrary grow size */;

		num_to_alloc = (yy_buffer_stack_max) + grow_size;
		(yy_buffer_stack) = (struct yy_buffer_state**)yyrealloc
								((yy_buffer_stack),
								num_to_alloc * sizeof(struct yy_buffer_state*)
								);
		if ( ! (yy_buffer_stack) )
			YY_FATAL_ERROR( "out of dynamic memory in yyensure_buffer_stack()" );

		/* zero only the new slots.*/
		memset((yy_buffer_stack) + (yy_buffer_stack_max), 0, grow_size * sizeof(struct yy_buffer_state*));
		(yy_buffer_stack_max) = num_to_alloc;
	}
}
/* %endif */

/* %if-c-only */
/** Setup the input buffer state to scan directly from a user-specified character buffer.
 * @param base the character buffer
 * @param size the size in bytes of the character buffer
 * 
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE yy_scan_buffer  (char * base, yy_size_t  size )
{
	YY_BUFFER_STATE b;
    
	if ( size < 2 ||
	     base[size-2] != YY_END_OF_BUFFER_CHAR ||
	     base[size-1] != YY_END_OF_BUFFER_CHAR )
		/* They forgot to leave room for the EOB's. */
		return NULL;

	b = (YY_BUFFER_STATE) yyalloc( sizeof( struct yy_buffer_state )  );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in yy_scan_buffer()" );

	b->yy_buf_size = (int) (size - 2);	/* "- 2" to take care of EOB's */
	b->yy_buf_pos = b->yy_ch_buf = base;
	b->yy_is_our_buffer = 0;
	b->yy_input_file = NULL;
	b->yy_n_chars = b->yy_buf_size;
	b->yy_is_interactive = 0;
	b->yy_at_bol = 1;
	b->yy_fill_buffer = 0;
	b->yy_buffer_status = YY_BUFFER_NEW;

	yy_switch_to_buffer( b  );

	return b;
}
/* %endif */

/* %if-c-only */
/** Setup the input buffer state to scan a string. The next call to yylex() will
 * scan from a @e copy of @a str.
 * @param yystr a NUL-terminated string to scan
 * 
 * @return the newly allocated buffer state object.
 * @note If you want to scan bytes that may contain NUL values, then use
 *       yy_scan_bytes() instead.
 */
YY_BUFFER_STATE yy_scan_string (const char * yystr )
{
    
	return yy_scan_bytes( yystr, (int) strlen(yystr) );
}
/* %endif */

/* %if-c-only */
/** Setup the input buffer state to scan the given bytes. The next call to yylex() will
 * scan from a @e copy of @a bytes.
 * @param yybytes the byte buffer to scan
 * @param _yybytes_len the number of bytes in the buffer pointed to by @a bytes.
 * 
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE yy_scan_bytes  (const char * yybytes, int  _yybytes_len )
{
	YY_BUFFER_STATE b;
	char *buf;
	yy_size_t n;
	int i;
    
	/* Get memory for full buffer, including space for trailing EOB's. */
	n = (yy_size_t) (_yybytes_len + 2);
	buf = (char *) yyalloc( n  );
	if ( ! buf )
		YY_FATAL_ERROR( "out of dynamic memory in yy_scan_bytes()" );

	for ( i = 0; i < _yybytes_len; ++i )
		buf[i] = yybytes[i];

	buf[_yybytes_len] = buf[_yybytes_len+1] = YY_END_OF_BUFFER_CHAR;

	b = yy_scan_buffer( buf, n );
	if ( ! b )
		YY_FATAL_ERROR( "bad buffer in yy_scan_bytes()" );

	/* It's okay to grow etc. this buffer, and we should throw it
	 * away when we're done.
	 */
	b->yy_is_our_buffer = 1;

	return b;
}
/* %endif */

/* %if-c-only */
    static void yy_push_state (int  _new_state )
/* %endif */
/* %if-c++-only */
/* %endif */
{
    	if ( (yy_start_stack_ptr) >= (yy_start_stack_depth) )
		{
		yy_size_t new_size;

		(yy_start_stack_depth) += YY_START_STACK_INCR;
		new_size = (yy_size_t) (yy_start_stack_depth) * sizeof( int );

		if ( ! (yy_start_stack) )
			(yy_start_stack) = (int *) yyalloc( new_size  );

		else
			(yy_start_stack) = (int *) yyrealloc(
					(void *) (yy_start_stack), new_size  );

		if ( ! (yy_start_stack) )
			YY_FATAL_ERROR( "out of memory expanding start-condition stack" );
		}

	(yy_start_stack)[(yy_start_stack_ptr)++] = YY_START;

	BEGIN(_new_state);
}

/* %if-c-only */
    static void yy_pop_state  (void)
/* %endif */
/* %if-c++-only */
/* %endif */
{
    	if ( --(yy_start_stack_ptr) < 0 )
		YY_FATAL_ERROR( "start-condition stack underflow" );

	BEGIN((yy_start_stack)[(yy_start_stack_ptr)]);
}

/* %if-c-only */
    static int yy_top_state  (void)
/* %endif */
/* %if-c++-only */
/* %endif */
{
    	return (yy_start_stack)[(yy_start_stack_ptr) - 1];
}

#ifndef YY_EXIT_FAILURE
#define YY_EXIT_FAILURE 2
#endif

/* %if-c-only */
static void yynoreturn yy_fatal_error (const char* msg )
{
			fprintf( stderr, "%s\n", msg );
	exit( YY_EXIT_FAILURE );
}
/* %endif */
/* %if-c++-only */
/* %endif */

/* Redefine yyless() so it works in section 3 code. */

#undef yyless
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		yytext[yyleng] = (yy_hold_char); \
		(yy_c_buf_p) = yytext + yyless_macro_arg; \
		(yy_hold_char) = *(yy_c_buf_p); \
		*(yy_c_buf_p) = '\0'; \
		yyleng = yyless_macro_arg; \
		} \
	while ( 0 )

/* Accessor  methods (get/set functions) to struct members. */

/* %if-c-only */
/* %if-reentrant */
/* %endif */

/** Get the current line number.
 * 
 */
int yyget_lineno  (void)
{
    
    return yylineno;
}

/** Get the input stream.
 * 
 */
FILE *yyget_in  (void)
{
        return yyin;
}

/** Get the output stream.
 * 
 */
FILE *yyget_out  (void)
{
        return yyout;
}

/** Get the length of the current token.
 * 
 */
int yyget_leng  (void)
{
        return yyleng;
}

/** Get the current token.
 * 
 */

char *yyget_text  (void)
{
        return yytext;
}

/* %if-reentrant */
/* %endif */

/** Set the current line number.
 * @param _line_number line number
 * 
 */
void yyset_lineno (int  _line_number )
{
    
    yylineno = _line_number;
}

/** Set the input stream. This does not discard the current
 * input buffer.
 * @param _in_str A readable stream.
 * 
 * @see yy_switch_to_buffer
 */
void yyset_in (FILE *  _in_str )
{
        yyin = _in_str ;
}

void yyset_out (FILE *  _out_str )
{
        yyout = _out_str ;
}

int yyget_debug  (void)
{
        return yy_flex_debug;
}

void yyset_debug (int  _bdebug )
{
        yy_flex_debug = _bdebug ;
}

/* %endif */

/* %if-reentrant */
/* %if-bison-bridge */
/* %endif */
/* %endif if-c-only */

/* %if-c-only */
static int yy_init_globals (void)
{
        /* Initialization is the same as for the non-reentrant scanner.
     * This function is called from yylex_destroy(), so don't allocate here.
     */

    (yy_buffer_stack) = NULL;
    (yy_buffer_stack_top) = 0;
    (yy_buffer_stack_max) = 0;
    (yy_c_buf_p) = NULL;
    (yy_init) = 0;
    (yy_start) = 0;

    (yy_start_stack_ptr) = 0;
    (yy_start_stack_depth) = 0;
    (yy_start_stack) =  NULL;

/* Defined in main.c */
#ifdef YY_STDINIT
    yyin = stdin;
    yyout = stdout;
#else
    yyin = NULL;
    yyout = NULL;
#endif

    /* For future reference: Set errno on error, since we are called by
     * yylex_init()
     */
    return 0;
}
/* %endif */

/* %if-c-only SNIP! this currently causes conflicts with the c++ scanner */
/* yylex_destroy is for both reentrant and non-reentrant scanners. */
int yylex_destroy  (void)
{
    
    /* Pop the buffer stack, destroying each element. */
	while(YY_CURRENT_BUFFER){
		yy_delete_buffer( YY_CURRENT_BUFFER  );
		YY_CURRENT_BUFFER_LVALUE = NULL;
		yypop_buffer_state();
	}

	/* Destroy the stack itself. */
	yyfree((yy_buffer_stack) );
	(yy_buffer_stack) = NULL;

    /* Destroy the start condition stack. */
        yyfree( (yy_start_stack)  );
        (yy_start_stack) = NULL;

    /* Reset the globals. This is important in a non-reentrant scanner so the next time
     * yylex() is called, initialization will occur. */
    yy_init_globals( );

/* %if-reentrant */
/* %endif */
    return 0;
}
/* %endif */

/*
 * Internal utility routines.
 */

#ifndef yytext_ptr
static void yy_flex_strncpy (char* s1, const char * s2, int n )
{
		
	int i;
	for ( i = 0; i < n; ++i )
		s1[i] = s2[i];
}
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen (const char * s )
{
	int n;
	for ( n = 0; s[n]; ++n )
		;

	return n;
}
#endif

void *yyalloc (yy_size_t  size )
{
			return malloc(size);
}

void *yyrealloc  (void * ptr, yy_size_t  size )
{
		
	/* The cast to (char *) in the following accommodates both
	 * implementations that use char* generic pointers, and those
	 * that use void* generic pointers.  It works with the latter
	 * because both ANSI C and C++ allow castless assignment from
	 * any pointer type to void*, and deal with argument conversions
	 * as though doing an assignment.
	 */
	return realloc(ptr, size);
}

void yyfree (void * ptr )
{
			free( (char *) ptr );	/* see yyrealloc() for (char *) cast */
}

/* %if-tables-serialization definitions */
/* %define-yytables   The name for this specific scanner's tables. */
#define YYTABLES_NAME "yytables"
/* %endif */

/* %ok-for-header */

#line 250 "lexer.l"




void
driver::scan_begin (const std::string& content)
{
  yy_flex_debug = trace_scanning;
  yy_scan_string(content.c_str());
}


void
driver::scan_end ()
{
  yy_delete_buffer(YY_CURRENT_BUFFER);
}

using namespace yy;

std::map<std::string, std::function<yy::parser::symbol_type(const yy::parser::location_type&)>> pragmas = {
    {"incoherent", yy::parser::make_INCOHERENT_PRAG},
    {"inline"      , yy::parser::make_INLINE_PRAG},
    {"overlappable", yy::parser::make_OVERLAPPABLE_PRAG},
    {"overlapping", yy::parser::make_OVERLAPPING_PRAG},
    {"overlaps", yy::parser::make_OVERLAPS_PRAG},
    {"specialize", yy::parser::make_SPECIALIZE_PRAG}
};

std::string clean_pragma(std::string prag)
{
    // 1. Remove pragma prefix
    if (prag.starts_with("{-#"))
        prag = prag.substr(3);

    // 2. Lower-case all the characters
    for(auto& c: prag)
        c = std::tolower(c);

    // 3. Split into words
    std::vector<std::string> words;
    bool inword = false;
    for(auto c: prag)
    {
        if (std::isspace(c))
            inword = false;
        else
        {
            if (not inword)
            {
                words.push_back({});
                inword=true;
            }
            words.back() += c;
        }
    }
    
    // 4. Connect the words with ' '
    return join(words, ' ');
}

std::optional<yy::parser::symbol_type> driver::prag(const yy::parser::location_type& loc)
{
    auto prag = clean_pragma(yytext);
    auto it = pragmas.find(prag);
    if (it == pragmas.end())
    {
        push_warning_message(loc, "Unknown pragma");
        return {};
    }
    else
        return it->second(loc);
}

yy::parser::symbol_type driver::varid(const yy::parser::location_type& loc) const
{
    auto it = reserved_words.find(yytext);
    if (it == reserved_words.end())
	return yy::parser::make_VARID  (yytext, loc);
    else
    {
	auto tok = it->second.first;
	switch(tok)
	{
	case parser::token::TOK_DO:
	case parser::token::TOK_MDO:
//	    std::cerr<<"switch to <layout_do> @ "<<loc<<"\n";
	    yy_push_state(layout_do);
	    break;
	case parser::token::TOK_OF:
	case parser::token::TOK_LCASE:
	case parser::token::TOK_LET:
	case parser::token::TOK_WHERE:
	case parser::token::TOK_REC:
//	    std::cerr<<"switch to <layout> @ "<<loc<<"\n";
	    yy_push_state(layout);
	    break;
	case parser::token::TOK_IF:
//	    std::cerr<<"switch to <layout_do> @ "<<loc<<"\n";
	    yy_push_state(layout_if);
	    break;
	default:
	    break;
	}
	return yy::parser::symbol_type(tok, loc);
    }
}

yy::parser::symbol_type driver::varsym(bool precededByClosing, bool followedByOpening, const yy::parser::location_type& loc) const
{
    // if loose infix occurrences are always operators, then:
    // if we handle type operators, then we should be able to parse (a ~ b) with ~ being a type operator.

    // tight infix reserved symbols
    if (precededByClosing and followedByOpening)
    {
        auto it = tight_infix_reserved_symbols.find(yytext);
        if (it != tight_infix_reserved_symbols.end())
        {
            auto tok = it->second.first;
            return yy::parser::symbol_type(tok, loc);
        }
    }

    // prefix reserved symbols
    if (precededByClosing == false and followedByOpening == true)
    {
        auto it = prefix_reserved_symbols.find(yytext);
        if (it != prefix_reserved_symbols.end())
        {
            auto tok = it->second.first;
            return yy::parser::symbol_type(tok, loc);
        }
    }

    // other reserved symbols
    auto it = reserved_symbols.find(yytext);
    if (it == reserved_symbols.end())
	return yy::parser::make_VARSYM  (yytext, loc);

    // general varsyms
    auto tok = it->second.first;
    return yy::parser::symbol_type(tok, loc);
}

yy::parser::symbol_type driver::consym(const yy::parser::location_type& loc) const
{
    auto it = reserved_symbols.find(yytext);
    if (it == reserved_symbols.end())
	return yy::parser::make_CONSYM  (yytext, loc);
    else
    {
	auto tok = it->second.first;
	return yy::parser::symbol_type(tok, loc);
    }
}

yy::parser::symbol_type make_integer10(const yy::parser::location_type& loc)
{
    return yy::parser::make_INTEGER(Haskell::integerFromString(yytext), loc);
}

yy::parser::symbol_type make_boxed_integer10(const yy::parser::location_type& loc)
{
    std::string s(yytext);
    assert(s.back() == '#');
    s.resize(s.size()-1);
    return yy::parser::make_PRIMINTEGER(Haskell::integerFromString(s), loc);
}

yy::parser::symbol_type make_integer(int radix, bool positive, int shift,
				     const yy::parser::location_type& loc)
{
    // This doesn't handle underscore separators
    errno = 0;
    auto n = std::stoi(yytext+shift, NULL, radix);
    if ( not (INT_MIN <= n and n <= INT_MAX and errno != ERANGE))
	throw yy::parser::syntax_error(loc, "integer is out of range: " + std::string(yytext));
    if (not positive)
	n = -n;
    return yy::parser::make_INTEGER(n, loc);
}

yy::parser::symbol_type make_boxed_integer(int radix, bool positive, int shift,
                                           const yy::parser::location_type& loc)
{
    // This doesn't handle underscore separators
    errno = 0;
    auto n = std::stoi(yytext+shift, NULL, radix);
    if ( not (INT_MIN <= n and n <= INT_MAX and errno != ERANGE))
	throw yy::parser::syntax_error(loc, "integer is out of range: " + std::string(yytext));
    if (not positive)
	n = -n;
    return yy::parser::make_PRIMINTEGER(n, loc);
}

#include "computation/haskell/literal.H"

yy::parser::symbol_type make_rational(const yy::parser::location_type& loc)
{
    rational r = Haskell::rationalFromString(yytext);
    return yy::parser::make_RATIONAL(r, loc);
}

std::optional<char> get_char(const char* s, int& i)
{
    char c = s[i];
    if (c == '\\')
    {
	i++;
	c = s[i];
	if (c == 'a')
	    c = '\a';
	else if (c == 'b')
	    c = '\b';
	else if (c == 'f')
	    c = '\f';
	else if (c == 'n')
	    c = '\n';
	else if (c == 'r')
	    c = '\r';
	else if (c == 't')
	    c = '\t';
	else if (c == 'v')
	    c = '\v';
	else if (c == '"')
	    c = '"';
	else if (c == '\'')
	    c = '\'';
	else if (c == '\\')
	    c = '\\';
	else
            return {};
    }
    return c;
}

yy::parser::symbol_type make_char(const yy::parser::location_type& loc, driver& drv)
{
    int i=1;
    auto c = get_char(yytext, i);

    // Check for unconsumed "\" right before the final "'"
    if (i == yyleng - 1)
    {
        assert(yyleng > 2 and yytext[yyleng-2] == '\\');
        yy::parser::location_type loc2(loc.begin+(yyleng-2));
        loc2 += 2;
        auto msg = Note()<<"invalid escape of ending quote";
        drv.push_error_message(loc2, msg);

        return yy::parser::make_CHAR('?',loc);
    }

    if (c)
        return yy::parser::make_CHAR(*c,loc);
    else
    {
        yy::parser::location_type loc2(loc.begin+1);
        loc2 += 2;
        auto msg = Note()<<"invalid escape sequence '"<<yytext[1]<<yytext[2]<<"'";
        drv.push_error_message(loc2, msg);

        return yy::parser::make_CHAR('?',loc);
    }
            
}

yy::parser::symbol_type make_string(const yy::parser::location_type& loc, driver& drv)
{
    std::string s;
    s.resize(yyleng);
    int j=0;
    for(int i=1;i<yyleng-1;i++)
    {
        if (auto c = get_char(yytext, i))
            s[j++] = *c;
        else
        {
            yy::parser::location_type loc2(loc.begin+(i-1));
            loc2 += 2;
            auto msg = Note()<<"invalid escape sequence '"<<yytext[i-1]<<yytext[i]<<"'";
            drv.push_error_message(loc2, msg);

            s[j++] = '?';
        }

        // Check for unconsumed "\" right before the final "'"
        if (i == yyleng - 1)
        {
            assert(yyleng > 2 and yytext[yyleng-2] == '\\');
            yy::parser::location_type loc2(loc.begin+(yyleng-2));
            loc2 += 2;
            auto msg = Note()<<"invalid escape of ending quote";
            drv.push_error_message(loc2, msg);

            s[j-1] = '?';
        }
    }
    s.resize(j);

    return yy::parser::make_STRING(s,loc);
}

std::optional<yy::parser::symbol_type> driver::do_bol(const yy::parser::location_type& loc)
{
    auto x = get_offside(loc);
    int delta_offset = x.offset;
    bool gen_semis = x.gen_semis;
//    std::cerr<<"do_bol @ "<<loc<<std::endl;
    if (delta_offset < 0)
    {
//	std::cerr<<"offset = "<<loc.end.column<<", LT, generating '}'"<<std::endl;
	pop_context();
	return yy::parser::make_VCCURLY(loc);
    }
    else if (delta_offset == 0 and gen_semis)
    {
//	std::cerr<<"offset = "<<loc.end.column<<", EQ, generating ';'"<<std::endl;
	yy_pop_state();
	return yy::parser::make_SEMI(loc);
    }
    else
    {
	yy_pop_state();
	return {};
    }
}

yy::parser::symbol_type driver::new_layout_context(const yy::parser::location_type& loc, bool strict, bool gen_semis, yy::parser::token_type tok)
{
    yy_pop_state();
    int offset = loc.end.column;
//    std::cerr<<"Opening new layout context at "<<loc<<std::endl;
    if (auto layout_context = get_context())
    {
	int prev_offset = layout_context->offset;
	if ((strict and prev_offset > offset) or
	    ((not strict) and prev_offset >= offset))
	{
//	    std::cerr<<"   - Going to layout_left"<<std::endl;
	    yy_push_state(layout_left);
	    return yy::parser::symbol_type(tok,loc);
	}
    }

    push_context({offset,gen_semis});
    return yy::parser::symbol_type(tok,loc);
}

yy::parser::symbol_type driver::do_layout_left(const yy::parser::location_type& loc)
{
//    std::cerr<<"Doing layout_left at "<<loc<<std::endl;
    yy_pop_state();
    yy_push_state(bol);
    return yy::parser::make_VCCURLY(loc);
}

yy::parser::symbol_type driver::hopefully_open_brace(const yy::parser::location_type& loc)
{
    if (auto layout_context = get_context())
    {
	int offset = loc.end.column;
	if (layout_context->offset >= offset)
	    throw yy::parser::syntax_error(loc,"Missing block");

    }
    yy_pop_state();
    push_context();
    return yy::parser::make_OCURLY(loc);
}

void use_things() {
    yy_top_state();
}

